# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.symbolics import *
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])


def testReduceOperations():

    f_1 = format["float"](1)
    f_2 = format["float"](2)

    # Aux. variables
    f2 = FloatValue(2)
    f0_5 = FloatValue(0.5)
    f1 = FloatValue(1.0)
    fm1 = FloatValue(-1.0)

    x = Symbol("x", GEO)
    y = Symbol("y", GEO)
    z = Symbol("z", GEO)
    a = Symbol("a", GEO)
    b = Symbol("b", GEO)
    c = Symbol("c", GEO)
    d = Symbol("d", GEO)

    # Simple expand and reduce simple float and symbol objects
    fx2 = f2.expand()
    xx = x.expand()

    fr2 = fx2.reduce_ops()
    xr = xx.reduce_ops()

    assert f2 == fr2
    assert x == xr

    # Test product
    p0 = f2 * x
    p1 = y * x
    p2 = x * f2 / y
    p3 = x * Sum([x, y])

    px0 = p0.expand()
    px1 = p1.expand()

    pr0 = px0.reduce_ops()
    pr1 = px1.reduce_ops()

    assert p0 == pr0
    assert p1 == pr1

    # Test fraction
    F0 = Fraction(p0, y)
    F1 = Fraction(x, p0)
    F2 = Fraction(p0, p1)
    F3 = Fraction(Sum([x * x, x * y]), y)
    F4 = Fraction(Sum([f2 * x, x * y]), a)

    Fx0 = F0.expand()
    Fx1 = F1.expand()
    Fx2 = F2.expand()
    Fx3 = F3.expand()
    Fx4 = F4.expand()

    Fr0 = Fx0.reduce_ops()
    Fr1 = Fx1.reduce_ops()
    Fr2 = Fx2.reduce_ops()
    Fr3 = Fx3.reduce_ops()
    Fr4 = Fx4.reduce_ops()

    assert Fr0 == F0
    assert Fr1 == f0_5
    assert Fr2 == Fraction(f2, y)
    assert str(Fr3) == "x*(%s + x/y)" % f_1
    assert str(Fr4) == "x*(%s + y)/a" % f_2

    # Test sum
    # TODO: Here we might have to add additional tests
    S0 = Sum([x, y])
    S1 = Sum([p0, p1])
    S2 = Sum([x, p1])
    S3 = Sum([p0, f2 * y])
    S4 = Sum([f2 * p1, z * p1])
    S5 = Sum([x, x * x, x * x * x])
    S6 = Sum([a * x * x, b * x * x * x, c * x * x, d * x * x * x])
    S7 = Sum([p0, p1, x * x, f2 * z, y * z])
    S8 = Sum([a * y, b * y, x * x * x * y, x * x * x * z])
    S9 = Sum([a * y, b * y, c * y, x * x * x * y, f2 * x * x, x * x * x * z])
    S10 = Sum([f2 * x * x * y, x * x * y * z])
    S11 = Sum([f2 * x * x * y * y, x * x * y * y * z])
    S12 = Sum([f2 * x * x * y * y, x * x * y * y * z, a * z, b * z, c * z])
    S13 = Sum([Fraction(f1, x), Fraction(f1, y)])
    S14 = Sum([Fraction(fm1, x), Fraction(fm1, y)])
    S15 = Sum([Fraction(f2, x), Fraction(f2, x)])
    S16 = Sum([Fraction(f2 * x, y * z), Fraction(f0_5, y * z)])
    S17 = Sum([(f2 * x * y) / a, (x * y * z) / b])
    S18 = Sum([(x * y) / a, (x * z) / a, f2 / a, (f2 * x * y) / a])
    S19 = Sum([(f2 * x) / a, (x * y) / a, z * x])
    S20 = Product([Sum([x, y]), Fraction(a, b), Fraction(Product([c, d]), z)])
    S21 = Sum([a * x, b * x, c * x, x * y, x * z, f2 * y, a * y, b * y, f2 * z, a * z, b * z])
    S22 = Sum([FloatValue(0.5) * x / y, FloatValue(-0.5) * x / y])
    S23 = Sum([x * y * z, x * y * y * y * z * z * z, y * y * y * z * z * z * z, z * z * z * z * z])

    Sx0 = S0.expand()
    Sx1 = S1.expand()
    Sx2 = S2.expand()
    Sx3 = S3.expand()
    Sx4 = S4.expand()
    Sx5 = S5.expand()
    Sx6 = S6.expand()
    Sx7 = S7.expand()
    Sx8 = S8.expand()
    Sx9 = S9.expand()
    Sx10 = S10.expand()
    Sx11 = S11.expand()
    Sx12 = S12.expand()
    Sx13 = S13.expand()
    Sx14 = S14.expand()
    Sx15 = S15.expand()
    Sx16 = S16.expand()
    Sx17 = S17.expand()
    Sx18 = S18.expand()
    Sx19 = S19.expand()
    Sx20 = S20.expand()
    Sx21 = S21.expand()
    Sx22 = S22.expand()
    Sx23 = S23.expand()

    Sr0 = Sx0.reduce_ops()
    Sr1 = Sx1.reduce_ops()
    Sr2 = Sx2.reduce_ops()
    Sr3 = Sx3.reduce_ops()
    Sr4 = Sx4.reduce_ops()
    Sr5 = Sx5.reduce_ops()
    Sr6 = Sx6.reduce_ops()
    Sr7 = Sx7.reduce_ops()
    Sr8 = Sx8.reduce_ops()
    Sr9 = Sx9.reduce_ops()
    Sr10 = Sx10.reduce_ops()
    Sr11 = Sx11.reduce_ops()
    Sr12 = Sx12.reduce_ops()
    Sr13 = Sx13.reduce_ops()
    Sr14 = Sx14.reduce_ops()
    Sr15 = Sx15.reduce_ops()
    Sr16 = Sx16.reduce_ops()
    Sr17 = Sx17.reduce_ops()
    Sr18 = Sx18.reduce_ops()
    Sr19 = Sx19.reduce_ops()
    Sr20 = Sx20.reduce_ops()
    Sr21 = Sx21.reduce_ops()
    Sr22 = Sx22.reduce_ops()
    Sr23 = Sx23.reduce_ops()

    assert Sr0 == S0
    assert str(Sr1) == "x*(%s + y)" % f_2
    # TODO: Should this be (x + x*y)?
    assert str(Sr2) == "x*(%s + y)" % f_1
    # assert str(Sr2), "(x + x*y)")
    assert str(Sr3) == "%s*(x + y)" % f_2
    assert str(Sr4) == "x*y*(%s + z)" % f_2
    assert str(Sr5) == "x*(%s + x*(%s + x))" % (f_1, f_1)
    assert str(Sr6) == "x*x*(a + c + x*(b + d))"
    assert str(Sr7) == "(x*(%s + x + y) + z*(%s + y))" % (f_2, f_2)
    assert str(Sr8) == "(x*x*x*(y + z) + y*(a + b))"
    assert str(Sr9) == "(x*x*(%s + x*(y + z)) + y*(a + b + c))" % f_2
    assert str(Sr10) == "x*x*y*(%s + z)" % f_2
    assert str(Sr11) == "x*x*y*y*(%s + z)" % f_2
    assert str(Sr12) == "(x*x*y*y*(%s + z) + z*(a + b + c))" % f_2
    assert str(Sr13) == "(%s/x + %s/y)" % (f_1, f_1)
    assert str(Sr14) == "(-%s/x-%s/y)" % (f_1, f_1)
    assert str(Sr15) == "%s/x" % format["float"](4)
    assert str(Sr16) == "(%s + %s*x)/(y*z)" % (format["float"](0.5), f_2)
    assert str(Sr17) == "x*y*(%s/a + z/b)" % f_2
    assert str(Sr18) == "(%s + x*(z + %s*y))/a" % (f_2, format["float"](3))
    assert str(Sr19) == "x*(z + (%s + y)/a)" % f_2
    assert str(Sr20) == "a*c*d*(x + y)/(b*z)"
    assert str(Sr21) == "(x*(a + b + c + y + z) + y*(%s + a + b) + z*(%s + a + b))" % (f_2, f_2)
    assert str(Sr22) == "%s" % format["float"](0)
    assert str(Sr23) == "(x*y*z + z*z*z*(y*y*y*(x + z) + z*z))"

    assert S0.ops() == 1
    assert Sr0.ops() == 1
    assert S1.ops() == 3
    assert Sr1.ops() == 2
    assert S2.ops() == 2
    assert Sr2.ops() == 2
    assert S3.ops() == 3
    assert Sr3.ops() == 2
    assert S4.ops() == 5
    assert Sr4.ops() == 3
    assert S5.ops() == 5
    assert Sr5.ops() == 4
    assert S6.ops() == 13
    assert Sr6.ops() == 6
    assert S7.ops() == 9
    assert Sr7.ops() == 6
    assert S8.ops() == 11
    assert Sr8.ops() == 7
    assert S9.ops() == 16
    assert Sr9.ops() == 9
    assert S10.ops() == 7
    assert Sr10.ops() == 4
    assert S11.ops() == 9
    assert Sr11.ops() == 5
    assert S12.ops() == 15
    assert Sr12.ops() == 9
    assert S13.ops() == 3
    assert Sr13.ops() == 3
    assert S14.ops() == 3
    assert Sr14.ops() == 3
    assert S15.ops() == 3
    assert Sr15.ops() == 1
    assert S16.ops() == 6
    assert Sr16.ops() == 4
    assert S17.ops() == 7
    assert Sr17.ops() == 5
    assert S18.ops() == 11
    assert Sr18.ops() == 5
    assert S19.ops() == 7
    assert Sr19.ops() == 4
    assert S20.ops() == 6
    assert Sr20.ops() == 6
    assert S21.ops() == 21
    assert Sr21.ops() == 13
    assert S23.ops() == 21
    assert Sr23.ops() == 12



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.reduce_operations import operation_count, expand_operations, reduce_operations
from ffc.quadrature.symbolics import *
from ffc.quadrature.sumobj import _group_fractions
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])
from ffc.log import error, push_level, pop_level, CRITICAL


def testSymbolOperators():
    "Test binary operators"

    f_0 = format["float"](0)
    f_1 = format["float"](1)
    f_2 = format["float"](2)
    f_3 = format["float"](3)
    f_0_5 = format["float"](0.5)
    f0 = FloatValue(0.0)
    f2 = FloatValue(2.0)
    fm1 = FloatValue(-1.0)
    fm3 = FloatValue(-3.0)

    x = Symbol("x", GEO)
    y = Symbol("y", GEO)
    z = Symbol("z", GEO)

    p0 = Product([f2, x])
    p1 = Product([x, y])
    p2 = Product([f2, z])
    p3 = Product([y, x, z])

    S0 = Sum([x, y])
    S1 = Sum([x, z])

    F0 = Fraction(f2, y)
    F1 = Fraction(x, y)
    F2 = Fraction(x, S0)
    F3 = Fraction(x, y)
    F4 = Fraction(p0, y)
    F5 = Fraction(fm3, y)

    # Test Symbol '+'
    assert str(x + f2) == '(%s + x)' % f_2
    assert str(x + x) == '%s*x' % f_2
    assert str(x + y) == '(x + y)'
    assert str(x + p0) == '%s*x' % f_3
    assert str(x + p1) == '(x + x*y)'
    assert str(x + S0) == '(x + x + y)'
    assert str(x + F0) == '(x + %s/y)' % f_2

    # Test Symbol '-'
    assert str(x - f2) == '(x-%s)' % f_2
    assert str(x - x) == '%s' % f_0
    assert str(x - y) == '(x - y)'
    assert str(x - p0) == ' - x'
    assert str(x - p1) == '(x - x*y)'
    assert str(x - S0) == '(x - (x + y))'
    assert str(x - F5) == '(x - -%s/y)' % f_3

    # Test Symbol '*', only need to test float, symbol and product. Sum and
    # fraction are handled by 'other'
    assert str(x * f2) == '%s*x' % f_2
    assert str(x * y) == 'x*y'
    assert str(x * p1) == 'x*x*y'

    # Test Symbol '/'
    assert str(x / f2) == '%s*x' % f_0_5
    assert str(x / x) == '%s' % f_1
    assert str(x / y) == 'x/y'
    assert str(x / S0) == 'x/(x + y)'
    assert str(x / p0) == '%s' % f_0_5
    assert str(y / p1) == '%s/x' % f_1
    assert str(z / p0) == '%s*z/x' % f_0_5
    assert str(z / p1) == 'z/(x*y)'
    with pytest.raises(Exception):
        truediv(x, F0)
    with pytest.raises(Exception):
        truediv(y, FloatValue(0))



# -*- coding: utf-8 -*-

# Copyright (C) 2005-2016 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Garth N. Wells, 2006-2009
# Modified by Marie E. Rognes, 2007-2010
# Modified by Kristian B. Oelgaard, 2010
# Modified by Lizao Li, 2015

# Python modules
import numpy

# FFC modules
from ffc.log import error

# UFL utils
from ufl.utils.sequences import product


class MixedElement:

    "Create a FFC mixed element from a list of FFC/FIAT elements."

    def __init__(self, elements):
        self._elements = elements
        self._entity_dofs = _combine_entity_dofs(self._elements)

    def elements(self):
        return self._elements

    def space_dimension(self):
        return sum(e.space_dimension() for e in self._elements)

    def value_shape(self):
        # Values of Tensor elements are flattened in MixedElements
        num_comps = lambda x: numpy.prod(x) if x else 1
        return (sum(num_comps(e.value_shape()) or 1 for e in self._elements),)

    def entity_dofs(self):
        return self._entity_dofs

    def mapping(self):
        return [m for e in self._elements for m in e.mapping()]

    def dual_basis(self):
        return [L for e in self._elements for L in e.dual_basis()]

    def num_components(self):
        return sum(_num_components(e) for e in self._elements)

    def tabulate(self, order, points):
        """
        Tabulate values on mixed element by appropriately reordering
        the tabulated values for the nested elements.

        The table of values is organized as follows:

          D^a v_i[j](x_k) = table[a][i][j][k]

        where a is a multi-index (tuple) representing the derivative.
        For example, a = (1, 1) represents d/dx d/dy.
        """

        # Special case: only one element
        # NOTE: KBO: Why do we need this special case? (FFC Bug #798578)
        #       When calling tabulate() on a MixedElement one should be able to
        #       rely on getting data back which is ordered like a mixed element
        #       irrespective of the number of elements?
#        if len(self._elements) == 1:
#            return self._elements[0].tabulate(order, points)

        # Zeros for insertion into mixed table
        table_shape = (self.space_dimension(), self.num_components(), len(points))

        # Iterate over elements and fill in non-zero values
        irange = (0, 0)
        crange = (0, 0)
        mixed_table = {}
        for element in self._elements:

            # Tabulate element
            table = element.tabulate(order, points)

            # Compute range for insertion into mixed table
            irange = (irange[1], irange[1] + element.space_dimension())
            crange = (crange[1], crange[1] + _num_components(element))

            # Insert table into mixed table
            for dtuple in table.keys():

                # Insert zeros if necessary (should only happen first time)
                if dtuple not in mixed_table:
                    # NOTE: It is super important to create a new numpy.zeros
                    # instance to avoid manipulating a numpy reference in case
                    # it is created outside the loop.
                    mixed_table[dtuple] = numpy.zeros(table_shape)

                # Insert non-zero values
                if (crange[1] - crange[0]) > 1:
                    mixed_table[dtuple][irange[0]:irange[1], crange[0]:crange[1]] = numpy.reshape(table[dtuple], (irange[1] - irange[0], crange[1] - crange[0], len(points)))
                else:
                    mixed_table[dtuple][irange[0]:irange[1], crange[0]] = table[dtuple]

        return mixed_table

#--- Utility functions ---


def _combine_entity_dofs(elements):
    """
    Combine the entity_dofs from a list of elements into a combined
    entity_dof containing the information for all the elements.
    """

    # Return {} if no elements
    if not elements:
        return {}

    # Initialize entity_dofs dictionary
    entity_dofs = dict((key, {}) for key in elements[0].entity_dofs())
    for dim in elements[0].entity_dofs():
        for entity in elements[0].entity_dofs()[dim]:
            entity_dofs[dim][entity] = []

    offset = 0

    # Insert dofs from each element into the mixed entity_dof.
    for e in elements:
        dofs = e.entity_dofs()
        for dim in dofs:
            for entity in dofs[dim]:

                # Must take offset into account
                shifted_dofs = [v + offset for v in dofs[dim][entity]]

                # Insert dofs from this element into the entity_dofs
                entity_dofs[dim][entity] += shifted_dofs

        # Adjust offset
        offset += e.space_dimension()
    return entity_dofs


def _num_components(element):
    "Compute number of components for element."
    return product(element.value_shape())



# -*- coding: utf-8 -*-

# Copyright (C) 2009-2016 Kristian B. Oelgaard and Martin Sandve Alnæs
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import numbers
import collections
import numpy

from ufl.sorting import sorted_expr_sum
from ufl import custom_integral_types
from ufl.classes import Integral
from ffc.representationutils import create_quadrature_points_and_weights


def collect_quadrature_rules(integrals, default_scheme, default_degree):
    "Collect quadrature rules found in list of integrals."
    rules = set()
    for integral in integrals:
        md = integral.metadata() or {}
        scheme = md.get("quadrature_rule", default_scheme)
        degree = md.get("quadrature_degree", default_degree)
        rule = (scheme, degree)
        rules.add(rule)
    return rules


def compute_quadrature_rules(rules, integral_type, cell):
    "Compute points and weights for a set of quadrature rules."
    quadrature_rules = {}
    quadrature_rule_sizes = {}
    for rule in rules:
        scheme, degree = rule

        # Compute quadrature points and weights
        (points, weights) = create_quadrature_points_and_weights(
            integral_type, cell, degree, scheme)

        if points is not None:
            points = numpy.asarray(points)

        if weights is None:
            # For custom integrals, there are no points
            num_points = None
        else:
            num_points = len(weights)

        # Assuming all rules with the same number of points are equal
        if num_points in quadrature_rules:
            assert quadrature_rules[num_points][0] == points
            assert quadrature_rules[num_points][0] == weights
            error("This number of points is already present in the weight table:\n  %s" % (quadrature_rules,))

        quadrature_rules[num_points] = (points, weights)
        quadrature_rule_sizes[rule] = num_points
    return quadrature_rules, quadrature_rule_sizes


def accumulate_integrals(itg_data, quadrature_rule_sizes):
    """Group and accumulate integrals according to the number
    of quadrature points in their rules.
    """
    if not itg_data.integrals:
        return {}

    # Group integrands by quadrature rule
    sorted_integrands = collections.defaultdict(list)
    if itg_data.integral_type in custom_integral_types:
        # Should only be one size here, ignoring irrelevant metadata and parameters
        num_points, = quadrature_rule_sizes.values()
        for integral in itg_data.integrals:
            sorted_integrands[num_points].append(integral.integrand())
    else:
        default_scheme = itg_data.metadata["quadrature_degree"]
        default_degree = itg_data.metadata["quadrature_rule"]
        for integral in itg_data.integrals:
            md = integral.metadata() or {}
            scheme = md.get("quadrature_rule", default_scheme)
            degree = md.get("quadrature_degree", default_degree)
            rule = (scheme, degree)
            num_points = quadrature_rule_sizes[rule]
            sorted_integrands[num_points].append(integral.integrand())

    # Accumulate integrands in a canonical ordering defined by UFL
    sorted_integrals = {
        num_points: Integral(
            sorted_expr_sum(integrands),
            itg_data.integral_type,
            itg_data.domain,
            itg_data.subdomain_id,
            {},
            None)
        for num_points, integrands in list(sorted_integrands.items())
        }
    return sorted_integrals



# -*- coding: utf-8 -*-


# TODO: Move these to uflacs.language utils?


def generate_return_new(L, classname, factory):
    if factory:
        return L.Return(L.Call("create_" + classname))
    else:
        return L.Return(L.New(classname))


def generate_return_new_switch(L, i, classnames, args=None, factory=False):
    if factory:
        def create(classname):
            return L.New(classname)
    else:
        def create(classname):
            return L.Call("create_" + classname)
    if classnames:
        cases = []
        if args is None:
            args = list(range(len(classnames)))
        for j, classname in zip(args, classnames):
            if classname:
                cases.append((j, L.Return(create(classname))))
        code = [L.Switch(i, cases, autobreak=False, autoscope=False)]
    else:
        code = []
    code.append(L.Return(L.Null()))
    return L.StatementList(code)



# -*- coding: utf-8 -*-
"QuadratureTransformer for quadrature code generation to translate UFL expressions."

# Copyright (C) 2009-2011 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Peter Brune 2009
# Modified by Anders Logg 2009, 2013
# Modified by Lizao Li, 2015, 2016

from six import iterkeys
from six.moves import xrange as range
from six import advance_iterator as next

# UFL common.
from ufl.utils.sorting import sorted_by_key
from ufl.measure import custom_integral_types, point_integral_types

# UFL Classes.
from ufl.classes import IntValue
from ufl.classes import FloatValue
from ufl.classes import Coefficient
from ufl.classes import Operator

# FFC modules.
from ffc.log import error, ffc_assert
from ffc.cpp import format

# Utility and optimisation functions for quadraturegenerator.
from ffc.quadrature.quadraturetransformerbase import QuadratureTransformerBase
from ffc.quadrature.quadratureutils import create_permutations
from ffc.quadrature.reduce_operations import operation_count
from ffc.quadrature.symbolics import IP


def firstkey(d):
    return next(iterkeys(d))


class QuadratureTransformer(QuadratureTransformerBase):

    "Transform UFL representation to quadrature code."

    def __init__(self, *args):

        # Initialise base class.
        QuadratureTransformerBase.__init__(self, *args)

    # -------------------------------------------------------------------------
    # Start handling UFL classes.
    # -------------------------------------------------------------------------
    # -------------------------------------------------------------------------
    # AlgebraOperators (algebra.py).
    # -------------------------------------------------------------------------
    def sum(self, o, *operands):
        # print("Visiting Sum: " + "\noperands: \n" + "\n".join(map(repr, operands)))

        # Prefetch formats to speed up code generation.
        f_group = format["grouping"]
        f_add = format["add"]
        f_mult = format["multiply"]
        f_float = format["floating point"]
        code = {}

        # Loop operands that has to be summed and sort according to map (j,k).
        for op in operands:
            # If entries does already exist we can add the code, otherwise just
            # dump them in the element tensor.
            for key, val in sorted_by_key(op):
                if key in code:
                    code[key].append(val)
                else:
                    code[key] = [val]

        # Add sums and group if necessary.
        for key, val in sorted_by_key(code):

            # Exclude all zero valued terms from sum
            value = [v for v in val if v is not None]

            if len(value) > 1:
                # NOTE: Since we no longer call expand_indices, the following
                # is needed to prevent the code from exploding for forms like
                # HyperElasticity
                duplications = {}
                for val in value:
                    if val in duplications:
                        duplications[val] += 1
                        continue
                    duplications[val] = 1

                # Add a product for each term that has duplicate code
                expressions = []
                for expr, num_occur in sorted_by_key(duplications):
                    if num_occur > 1:
                        # Pre-multiply expression with number of occurrences
                        expressions.append(f_mult([f_float(num_occur), expr]))
                        continue
                    # Just add expression if there is only one
                    expressions.append(expr)
                ffc_assert(expressions, "Where did the expressions go?")

                if len(expressions) > 1:
                    code[key] = f_group(f_add(expressions))
                    continue
                code[key] = expressions[0]
            else:
                # Check for zero valued sum and delete from code
                # This might result in returning an empty dict, but that should
                # be interpreted as zero by other handlers.
                if not value:
                    del code[key]
                    continue
                code[key] = value[0]

        return code

    def product(self, o, *operands):
        # print("Visiting Product with operands: \n" + "\n".join(map(repr,operands)))

        # Prefetch formats to speed up code generation.
        f_mult = format["multiply"]
        permute = []
        not_permute = []

        # Sort operands in objects that needs permutation and objects that does not.
        for op in operands:
            # If we get an empty dict, something was zero and so is the product.
            if not op:
                return {}
            if len(op) > 1 or (op and firstkey(op) != ()):
                permute.append(op)
            elif op and firstkey(op) == ():
                not_permute.append(op[()])

        # Create permutations.
        # print("\npermute: " + repr(permute))
        # print("\nnot_permute: " + repr(not_permute))
        permutations = create_permutations(permute)
        # print("\npermutations: " + repr(permutations))

        # Create code.
        code = {}
        if permutations:
            for key, val in sorted(permutations.items()):
                # Sort key in order to create a unique key.
                l = sorted(key)

                # Loop products, don't multiply by '1' and if we encounter a None the product is zero.
                # TODO: Need to find a way to remove and J_inv00 terms that might
                # disappear as a consequence of eliminating a zero valued term
                value = []
                zero = False
                for v in val + not_permute:
                    if v is None:
                        ffc_assert(tuple(l) not in code, "This key should not be in the code.")
                        code[tuple(l)] = None
                        zero = True
                        break
                    elif not v:
                        print("v: '%s'" % repr(v))
                        error("should not happen")
                    elif v == "1":
                        pass
                    else:
                        value.append(v)

                if not value:
                    value = ["1"]
                if zero:
                    code[tuple(l)] = None
                else:
                    code[tuple(l)] = f_mult(value)
        else:
            # Loop products, don't multiply by '1' and if we encounter a None the product is zero.
            # TODO: Need to find a way to remove terms from 'used sets' that might
            # disappear as a consequence of eliminating a zero valued term
            value = []
            for v in not_permute:
                if v is None:
                    code[()] = None
                    return code
                elif not v:
                    print("v: '%s'" % repr(v))
                    error("should not happen")
                elif v == "1":
                    pass
                else:
                    value.append(v)
            # We did have values, but they might have been all ones.
            if value == [] and not_permute != []:
                code[()] = f_mult(["1"])
            else:
                code[()] = f_mult(value)
        return code

    def division(self, o, *operands):
        # print("Visiting Division with operands: \n" + "\n".join(map(repr,operands)))

        # Prefetch formats to speed up code generation.
        f_div = format["div"]
        f_grouping = format["grouping"]

        ffc_assert(len(operands) == 2,
                   "Expected exactly two operands (numerator and denominator): " + repr(operands))

        # Get the code from the operands.
        numerator_code, denominator_code = operands

        # TODO: Are these safety checks needed? Need to check for None?
        ffc_assert(() in denominator_code and len(denominator_code) == 1,
                   "Only support function type denominator: " + repr(denominator_code))

        code = {}
        # Get denominator and create new values for the numerator.
        denominator = denominator_code[()]
        ffc_assert(denominator is not None, "Division by zero!")

        for key, val in numerator_code.items():
            # If numerator is None the fraction is also None
            if val is None:
                code[key] = None
            # If denominator is '1', just return numerator
            elif denominator == "1":
                code[key] = val
            # Create fraction and add to code
            else:
                code[key] = f_div(val, f_grouping(denominator))

        return code

    def power(self, o):
        # print("\n\nVisiting Power: " + repr(o))

        # Get base and exponent.
        base, expo = o.ufl_operands

        # Visit base to get base code.
        base_code = self.visit(base)

        # TODO: Are these safety checks needed? Need to check for None?
        ffc_assert(() in base_code and len(base_code) == 1, "Only support function type base: " + repr(base_code))

        # Get the base code.
        val = base_code[()]

        # Handle different exponents
        if isinstance(expo, IntValue):
            return {(): format["power"](val, expo.value())}
        elif isinstance(expo, FloatValue):
            return {(): format["std power"](val, format["floating point"](expo.value()))}
        elif isinstance(expo, (Coefficient, Operator)):
            exp = self.visit(expo)
            return {(): format["std power"](val, exp[()])}
        else:
            error("power does not support this exponent: " + repr(expo))

    def abs(self, o, *operands):
        # print("\n\nVisiting Abs: " + repr(o) + "with operands: " + "\n".join(map(repr,operands)))

        # Prefetch formats to speed up code generation.
        f_abs = format["absolute value"]

        # TODO: Are these safety checks needed? Need to check for None?
        ffc_assert(len(operands) == 1 and () in operands[0] and len(operands[0]) == 1,
                   "Abs expects one operand of function type: " + repr(operands))

        # Take absolute value of operand.
        return {(): f_abs(operands[0][()])}

    def min_value(self, o, *operands):
        f_min = format["min value"]
        return {(): f_min(operands[0][()], operands[1][()])}

    def max_value(self, o, *operands):
        f_max = format["max value"]
        return {(): f_max(operands[0][()], operands[1][()])}

    # -------------------------------------------------------------------------
    # Condition, Conditional (conditional.py).
    # -------------------------------------------------------------------------
    def not_condition(self, o, *operands):
        # This is a Condition but not a BinaryCondition, and the operand will be another Condition
        # Get condition expression and do safety checks.
        # Might be a bit too strict?
        cond, = operands
        ffc_assert(len(cond) == 1 and firstkey(cond) == (),
                   "Condition for NotCondition should only be one function: " + repr(cond))
        return {(): format["not"](cond[()])}

    def binary_condition(self, o, *operands):

        # Get LHS and RHS expressions and do safety checks.
        # Might be a bit too strict?
        lhs, rhs = operands
        ffc_assert(len(lhs) == 1 and firstkey(lhs) == (),
                   "LHS of Condition should only be one function: " + repr(lhs))
        ffc_assert(len(rhs) == 1 and firstkey(rhs) == (),
                   "RHS of Condition should only be one function: " + repr(rhs))

        # Map names from UFL to cpp.py.
        name_map = {"==": "is equal", "!=": "not equal",
                    "<": "less than", ">": "greater than",
                    "<=": "less equal", ">=": "greater equal",
                    "&&": "and", "||": "or"}

        # Get values and test for None
        l_val = lhs[()]
        r_val = rhs[()]
        if l_val is None:
            l_val = format["float"](0.0)
        if r_val is None:
            r_val = format["float"](0.0)

        return {(): format["grouping"](l_val + format[name_map[o._name]] + r_val)}

    def conditional(self, o, *operands):

        # Get condition and return values; and do safety check.
        cond, true, false = operands
        ffc_assert(len(cond) == 1 and firstkey(cond) == (),
                   "Condtion should only be one function: " + repr(cond))
        ffc_assert(len(true) == 1 and firstkey(true) == (),
                   "True value of Condtional should only be one function: " + repr(true))
        ffc_assert(len(false) == 1 and firstkey(false) == (),
                   "False value of Condtional should only be one function: " + repr(false))

        # Get values and test for None
        t_val = true[()]
        f_val = false[()]
        if t_val is None:
            t_val = format["float"](0.0)
        if f_val is None:
            f_val = format["float"](0.0)

        # Create expression for conditional
        expr = format["evaluate conditional"](cond[()], t_val, f_val)
        num = len(self.conditionals)
        name = format["conditional"](num)
        if expr not in self.conditionals:
            self.conditionals[expr] = (IP, operation_count(expr, format), num)
        else:
            num = self.conditionals[expr][2]
            name = format["conditional"](num)
        return {(): name}

    # -------------------------------------------------------------------------
    # FacetNormal, CellVolume, Circumradius, FacetArea (geometry.py).
    # -------------------------------------------------------------------------
    def cell_coordinate(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def facet_coordinate(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def cell_origin(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def facet_origin(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def cell_facet_origin(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def jacobian(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def jacobian_determinant(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def jacobian_inverse(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def facet_jacobian(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def facet_jacobian_determinant(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def facet_jacobian_inverse(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def cell_facet_jacobian(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def cell_facet_jacobian_determinant(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def cell_facet_jacobian_inverse(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def facet_normal(self, o):
        # print("Visiting FacetNormal:")

        # Get the component
        components = self.component()

        # Safety check.
        ffc_assert(len(components) == 1, "FacetNormal expects 1 component index: " + repr(components))

        # Handle 1D as a special case.
        # FIXME: KBO: This has to change for mD elements in R^n : m < n
        if self.gdim == 1:  # FIXME: MSA: UFL uses shape (1,) now, can we remove the special case here then?
            normal_component = format["normal component"](self.restriction, "")
        else:
            normal_component = format["normal component"](self.restriction, components[0])
        self.trans_set.add(normal_component)

        return {(): normal_component}

    def cell_normal(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def cell_volume(self, o):
        # FIXME: KBO: This has to change for higher order elements
        volume = format["cell volume"](self.restriction)
        self.trans_set.add(volume)

        return {(): volume}

    def circumradius(self, o):
        # FIXME: KBO: This has to change for higher order elements
        circumradius = format["circumradius"](self.restriction)
        self.trans_set.add(circumradius)

        return {(): circumradius}

    def facet_area(self, o):
        # FIXME: KBO: This has to change for higher order elements
        # NOTE: Omitting restriction because the area of a facet is the same
        # on both sides.
        # FIXME: Since we use the scale factor, facet area has no meaning
        # for cell integrals. (Need check in FFC or UFL).
        area = format["facet area"]
        self.trans_set.add(area)

        return {(): area}

    def min_facet_edge_length(self, o):
        # FIXME: this has no meaning for cell integrals. (Need check in FFC or UFL).

        tdim = self.tdim
        if tdim < 3:
            return self.facet_area(o)

        edgelen = format["min facet edge length"](self.restriction)
        self.trans_set.add(edgelen)

        return {(): edgelen}

    def max_facet_edge_length(self, o):
        # FIXME: this has no meaning for cell integrals. (Need check in FFC or UFL).

        tdim = self.tdim
        if tdim < 3:
            return self.facet_area(o)

        edgelen = format["max facet edge length"](self.restriction)
        self.trans_set.add(edgelen)

        return {(): edgelen}

    def cell_orientation(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    def quadrature_weight(self, o):  # FIXME
        error("This object should be implemented by the child class.")

    # -------------------------------------------------------------------------

    def create_argument(self, ufl_argument, derivatives, component, local_comp,
                        local_offset, ffc_element, transformation, multiindices,
                        tdim, gdim, avg):
        "Create code for basis functions, and update relevant tables of used basis."

        # Prefetch formats to speed up code generation.
        f_group = format["grouping"]
        f_add = format["add"]
        f_mult = format["multiply"]
        f_transform = format["transform"]
        f_detJ = format["det(J)"]
        f_inv = format["inverse"]

        # Reset code
        code = {}

        # Handle affine mappings.
        if transformation == "affine":

            # Loop derivatives and get multi indices.
            for multi in multiindices:
                deriv = [multi.count(i) for i in range(tdim)]
                if not any(deriv):
                    deriv = []

                # Create mapping and basis name.
                # print "component = ", component
                mapping, basis = self._create_mapping_basis(component, deriv,
                                                            avg, ufl_argument,
                                                            ffc_element)
                if mapping not in code:
                    code[mapping] = []

                if basis is not None:
                    # Add transformation
                    code[mapping].append(self.__apply_transform(basis,
                                                                derivatives,
                                                                multi, tdim,
                                                                gdim))

        # Handle non-affine mappings.
        else:
            ffc_assert(avg is None,
                       "Taking average is not supported for non-affine mappings.")

            # Loop derivatives and get multi indices.
            for multi in multiindices:
                deriv = [multi.count(i) for i in range(tdim)]
                if not any(deriv):
                    deriv = []

                if transformation in ["covariant piola",
                                      "contravariant piola"]:
                    for c in range(tdim):
                        # Create mapping and basis name.
                        mapping, basis = self._create_mapping_basis(c + local_offset, deriv, avg, ufl_argument, ffc_element)
                        if mapping not in code:
                            code[mapping] = []
                        if basis is not None:
                            # Multiply basis by appropriate transform.
                            if transformation == "covariant piola":
                                dxdX = f_transform("JINV", c, local_comp, tdim, gdim, self.restriction)
                                self.trans_set.add(dxdX)
                                basis = f_mult([dxdX, basis])
                            elif transformation == "contravariant piola":
                                self.trans_set.add(f_detJ(self.restriction))
                                detJ = f_inv(f_detJ(self.restriction))
                                dXdx = f_transform("J", local_comp, c, gdim, tdim, self.restriction)
                                self.trans_set.add(dXdx)
                                basis = f_mult([detJ, dXdx, basis])
                            # Add transformation if needed.
                            code[mapping].append(self.__apply_transform(basis, derivatives, multi, tdim, gdim))
                elif transformation == "double covariant piola":
                    # g_ij = (Jinv)_ki G_kl (Jinv)lj
                    i = local_comp // tdim
                    j = local_comp % tdim
                    for k in range(tdim):
                        for l in range(tdim):
                            # Create mapping and basis name.
                            mapping, basis = self._create_mapping_basis(
                                k * tdim + l + local_offset,
                                deriv, avg, ufl_argument, ffc_element)
                            if mapping not in code:
                                code[mapping] = []
                            if basis is not None:
                                J1 = f_transform("JINV", k, i, tdim, gdim,
                                                 self.restriction)
                                J2 = f_transform("JINV", l, j, tdim, gdim,
                                                 self.restriction)
                                self.trans_set.add(J1)
                                self.trans_set.add(J2)
                                basis = f_mult([J1, basis, J2])
                                # Add transformation if needed.
                                code[mapping].append(
                                    self.__apply_transform(
                                        basis, derivatives, multi,
                                        tdim, gdim))
                elif transformation == "double contravariant piola":
                    # g_ij = (detJ)^(-2) J_ik G_kl J_jl
                    i = local_comp // tdim
                    j = local_comp % tdim
                    for k in range(tdim):
                        for l in range(tdim):
                            # Create mapping and basis name.
                            mapping, basis = self._create_mapping_basis(
                                k * tdim + l + local_offset,
                                deriv, avg, ufl_argument, ffc_element)
                            if mapping not in code:
                                code[mapping] = []
                            if basis is not None:
                                J1 = f_transform("J", i, k, gdim, tdim,
                                                 self.restriction)
                                J2 = f_transform("J", j, l, gdim, tdim,
                                                 self.restriction)
                                self.trans_set.add(J1)
                                self.trans_set.add(J2)
                                self.trans_set.add(f_detJ(self.restriction))
                                invdetJ = f_inv(f_detJ(self.restriction))
                                basis = f_mult([invdetJ, invdetJ, J1, basis,
                                                J2])
                                # Add transformation if needed.
                                code[mapping].append(
                                    self.__apply_transform(
                                        basis, derivatives, multi,
                                        tdim, gdim))
                else:
                    error("Transformation is not supported: " +
                          repr(transformation))

        # Add sums and group if necessary.
        for key, val in list(code.items()):
            if len(val) > 1:
                code[key] = f_group(f_add(val))
            elif val:
                code[key] = val[0]
            else:
                # Return a None (zero) because val == []
                code[key] = None

        return code

    def create_function(self, ufl_function, derivatives, component, local_comp,
                        local_offset, ffc_element, is_quad_element,
                        transformation, multiindices, tdim, gdim, avg):
        "Create code for basis functions, and update relevant tables of used basis."
        ffc_assert(ufl_function in self._function_replace_values, "Expecting ufl_function to have been mapped prior to this call.")

        # Prefetch formats to speed up code generation.
        f_mult = format["multiply"]
        f_transform = format["transform"]
        f_detJ = format["det(J)"]
        f_inv = format["inverse"]

        # Reset code
        code = []

        # Handle affine mappings.
        if transformation == "affine":
            # Loop derivatives and get multi indices.
            for multi in multiindices:
                deriv = [multi.count(i) for i in range(tdim)]
                if not any(deriv):
                    deriv = []

                # Create function name.
                function_name = self._create_function_name(component, deriv, avg, is_quad_element, ufl_function, ffc_element)
                if function_name:
                    # Add transformation if needed.
                    code.append(self.__apply_transform(function_name, derivatives, multi, tdim, gdim))

        # Handle non-affine mappings.
        else:
            ffc_assert(avg is None, "Taking average is not supported for non-affine mappings.")

            # Loop derivatives and get multi indices.
            for multi in multiindices:
                deriv = [multi.count(i) for i in range(tdim)]
                if not any(deriv):
                    deriv = []

                if transformation in ["covariant piola", "contravariant piola"]:
                    # Vectors
                    for c in range(tdim):
                        function_name = self._create_function_name(c + local_offset, deriv, avg, is_quad_element, ufl_function, ffc_element)
                        if function_name:
                            # Multiply basis by appropriate transform.
                            if transformation == "covariant piola":
                                dxdX = f_transform("JINV", c, local_comp, tdim, gdim, self.restriction)
                                self.trans_set.add(dxdX)
                                function_name = f_mult([dxdX, function_name])
                            elif transformation == "contravariant piola":
                                self.trans_set.add(f_detJ(self.restriction))
                                detJ = f_inv(f_detJ(self.restriction))
                                dXdx = f_transform("J", local_comp, c, gdim, tdim, self.restriction)
                                self.trans_set.add(dXdx)
                                function_name = f_mult([detJ, dXdx, function_name])
                            else:
                                error("Transformation is not supported: ", repr(transformation))

                            # Add transformation if needed.
                            code.append(self.__apply_transform(function_name, derivatives, multi, tdim, gdim))
                elif transformation == "double covariant piola":
                    # g_ij = (Jinv)_ki G_kl (Jinv)lj
                    i = local_comp // tdim
                    j = local_comp % tdim
                    for k in range(tdim):
                        for l in range(tdim):
                            # Create mapping and basis name.
                            function_name = self._create_function_name(k * tdim + l + local_offset, deriv, avg, is_quad_element, ufl_function, ffc_element)
                            J1 = f_transform("JINV", k, i, tdim, gdim, self.restriction)
                            J2 = f_transform("JINV", l, j, tdim, gdim, self.restriction)
                            self.trans_set.add(J1)
                            self.trans_set.add(J2)
                            function_name = f_mult([J1, function_name, J2])
                            # Add transformation if needed.
                            code.append(self.__apply_transform(function_name, derivatives, multi, tdim, gdim))
                elif transformation == "double contravariant piola":
                    # g_ij = (detJ)^(-2) J_ik G_kl J_jl
                    i = local_comp // tdim
                    j = local_comp % tdim
                    for k in range(tdim):
                        for l in range(tdim):
                            # Create mapping and basis name.
                            function_name = self._create_function_name(
                                k * tdim + l + local_offset,
                                deriv, avg, is_quad_element,
                                ufl_function, ffc_element)
                            J1 = f_transform("J", i, k, tdim, gdim,
                                             self.restriction)
                            J2 = f_transform("J", j, l, tdim, gdim,
                                             self.restriction)
                            invdetJ = f_inv(f_detJ(self.restriction))
                            self.trans_set.add(J1)
                            self.trans_set.add(J2)
                            function_name = f_mult([invdetJ, invdetJ, J1,
                                                    function_name, J2])
                            # Add transformation if needed.
                            code.append(self.__apply_transform(function_name,
                                                               derivatives,
                                                               multi, tdim,
                                                               gdim))
                else:
                    error("Transformation is not supported: " + repr(transformation))

        if not code:
            return None
        elif len(code) > 1:
            code = format["grouping"](format["add"](code))
        else:
            code = code[0]

        return code

    # -------------------------------------------------------------------------
    # Helper functions for Argument and Coefficient
    # -------------------------------------------------------------------------
    def __apply_transform(self, function, derivatives, multi, tdim, gdim):  # XXX UFLACS REUSE
        "Apply transformation (from derivatives) to basis or function."
        f_transform = format["transform"]

        # Add transformation if needed.
        transforms = []
        if self.integral_type in custom_integral_types:
            for i, direction in enumerate(derivatives):
                # Custom integrals to not need transforms, so in place
                # of the transform, we insert an identity matrix
                ref = multi[i]
                if ref != direction:
                    transforms.append(0)

        else:
            for i, direction in enumerate(derivatives):
                ref = multi[i]
                t = f_transform("JINV", ref, direction, tdim, gdim, self.restriction)
                self.trans_set.add(t)
                transforms.append(t)

        # Only multiply by basis if it is present.
        if function:
            prods = transforms + [function]
        else:
            prods = transforms

        return format["multiply"](prods)

    # -------------------------------------------------------------------------
    # Helper functions for transformation of UFL objects in base class
    # -------------------------------------------------------------------------
    def _create_symbol(self, symbol, domain):
        return {(): symbol}

    def _create_product(self, symbols):
        return format["multiply"](symbols)

    def _format_scalar_value(self, value):
        # print("format_scalar_value: %d" % value)
        if value is None:
            return {(): None}
        # TODO: Handle value < 0 better such that we don't have + -2 in the code.
        return {(): format["floating point"](value)}

    def _math_function(self, operands, format_function):
        # TODO: Are these safety checks needed?
        ffc_assert(len(operands) == 1 and () in operands[0] and len(operands[0]) == 1,
                   "MathFunctions expect one operand of function type: " + repr(operands))
        # Use format function on value of operand.
        new_operand = {}
        operand = operands[0]
        for key, val in operand.items():
            new_operand[key] = format_function(val)
        return new_operand

    def _atan_2_function(self, operands, format_function):
        x1, x2 = operands
        x1, x2 = sorted(x1.values())[0], sorted(x2.values())[0]

        if x1 is None:
            x1 = format["floating point"](0.0)
        if x2 is None:
            x2 = format["floating point"](0.0)
        return {(): format_function(x1, x2)}

    def _bessel_function(self, operands, format_function):
        # TODO: Are these safety checks needed?
        ffc_assert(len(operands) == 2,
                   "BesselFunctions expect two operands of function type: " + repr(operands))
        nu, x = operands
        ffc_assert(len(nu) == 1 and () in nu,
                   "Expecting one operand of function type as first argument to BesselFunction : " + repr(nu))
        ffc_assert(len(x) == 1 and () in x,
                   "Expecting one operand of function type as second argument to BesselFunction : " + repr(x))
        nu = nu[()]
        x = x[()]
        if nu is None:
            nu = format["floating point"](0.0)
        if x is None:
            x = format["floating point"](0.0)

        # Use format function on arguments.
        # NOTE: Order of nu and x is reversed compared to the UFL and C++
        # function calls because of how Symbol treats exponents.
        # this will change once quadrature optimisations has been cleaned up.
        return {(): format_function(x, nu)}

    # -------------------------------------------------------------------------
    # Helper functions for code_generation()
    # -------------------------------------------------------------------------
    def _count_operations(self, expression):
        return operation_count(expression, format)

    def _create_entry_data(self, val, integral_type):
        # Multiply value by weight and determinant
        # Create weight and scale factor.

        weight = format["weight"](self.points)
        if self.points is None or self.points > 1:
            weight += format["component"]("", format["integration points"])

        # Update sets of used variables.
        if integral_type in (point_integral_types + custom_integral_types):
            trans_set = set()
            value = format["mul"]([val, weight])
        else:
            f_scale_factor = format["scale factor"]
            trans_set = set([f_scale_factor])
            value = format["mul"]([val, weight, f_scale_factor])

        trans_set.update(self.trans_set)
        used_points = set([self.points])
        ops = self._count_operations(value)
        used_psi_tables = set([v for k, v in self.psi_tables_map.items()])

        return (value, ops, [trans_set, used_points, used_psi_tables])



# -*- coding: utf-8 -*-
"This module provides simple C++ code for verification of UFC code."

# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

evaluate_basis_code_fiat = """\
#include <iostream>
#include <ufc.h>
#include <cstdlib>
#include "test.h"

int main(int argc, char* argv[])
{
  // Create element
  %(element)s element;

  // Get derivative order
  unsigned int n = std::atoi(argv[1]);

  // Value dimension
  int N;
  if (element.value_rank() == 0)
    N = 1;
  else
  {
    N = 1;
    for (unsigned int i = 0; i < element.value_rank(); i++)
      N = N * element.value_dimension(i);
  }

  // Compute number of derivatives.
  unsigned int  num_derivatives = 1;
  for (unsigned int r = 0; r < n; r++)
    num_derivatives *= %(dim)d;

  // Create values
  unsigned int num_dof_vals = N*num_derivatives;
  double* dof_values = new double[num_dof_vals];
  for (unsigned int i = 0; i < num_dof_vals; i++)
    dof_values[i] = 0.0;

  %(cell_ref_coords)s
  double coordinate_dofs[%(num_coords)d*%(dim)d];
  int k = 0;
  for (int i = 0; i < %(num_coords)d; i++)
  {
    for (int j = 0; j < %(dim)d; j++)
      coordinate_dofs[k++] = cell_ref_coords[i][j];
  }

  // Random points where we want to evaluate the basis functions
  // coordinates of dofs and three arbitrary points on the reference cell.
  double points[%(num_points)d][%(dim)d] = %(points)s

  // Init array of coordinates
  double coordinates[3] = {0,0,0};

  std::cout.precision(8);
  std::cout.setf(std::ios::fixed);

  // If we're testing evaluate_basis, loop all points.
  if (n == 0)
  {
    for (unsigned int p = 0; p < %(num_points)d; p++)
    {
      for (unsigned int d = 0; d < %(dim)d; d++)
        coordinates[d] = points[p][d];

      // Loop element space dimension and call evaluate_basis.
      for (unsigned int i = 0; i < element.space_dimension(); i++)
      {
        element.evaluate_basis(i, dof_values, coordinates, coordinate_dofs, 0);

        // Print values
        for (unsigned int j = 0; j < num_dof_vals; j++)
          std::cout << dof_values[j] << " ";
      }
      std::cout << std::endl;
    }
  }
  else
  {
    // Else loop the arbitrary 3 points, otherwise the number of tests explode
    // with the element.space_dimension()^2.
    for (unsigned int p = element.space_dimension(); p < %(num_points)d; p++)
    {
      for (unsigned int d = 0; d < %(dim)d; d++)
        coordinates[d] = points[p][d];

      // Loop element space dimension and call evaluate_basis.
      for (unsigned int i = 0; i < element.space_dimension(); i++)
      {
        element.evaluate_basis_derivatives(i, n, dof_values, coordinates,
                                           coordinate_dofs, 0);

        // Print values
        for (unsigned int j = 0; j < num_dof_vals; j++)
          std::cout << dof_values[j] << " ";
      }
      std::cout << std::endl;
    }
  }

  return 0;
}
"""



# -*- coding: utf-8 -*-
"This file contains functions to optimise the code generated for quadrature representation."

# Copyright (C) 2009-2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

from ufl.utils.sorting import sorted_by_key

# FFC modules
from ffc.log import error
from ffc.cpp import format

# TODO: Use proper errors, not just RuntimeError.
# TODO: Change all if value == 0.0 to something more safe.

# Some basic variables.
BASIS = 0
IP = 1
GEO = 2
CONST = 3
type_to_string = {BASIS: "BASIS", IP: "IP", GEO: "GEO", CONST: "CONST"}

# Functions and dictionaries for cache implementation.
# Increases speed and should also reduce memory consumption.
_float_cache = {}


def create_float(val):
    if val in _float_cache:
        return _float_cache[val]
    float_val = FloatValue(val)
    _float_cache[val] = float_val
    return float_val


_symbol_cache = {}


def create_symbol(variable, symbol_type, base_expr=None, base_op=0):
    key = (variable, symbol_type, base_expr, base_op)
    if key in _symbol_cache:
        return _symbol_cache[key]
    symbol = Symbol(variable, symbol_type, base_expr, base_op)
    _symbol_cache[key] = symbol
    return symbol


_product_cache = {}


def create_product(variables):
    # NOTE: If I switch on the sorted line, it might be possible to find more
    # variables in the cache, but it adds some overhead so I don't think it
    # pays off. The member variables are also sorted in the classes
    # (Product and Sum) so the list 'variables' is probably already sorted.
    key = tuple(variables)
    if key in _product_cache:
        return _product_cache[key]
    product = Product(key)
    _product_cache[key] = product
    return product


_sum_cache = {}


def create_sum(variables):
    # NOTE: If I switch on the sorted line, it might be possible to
    # find more variables in the cache, but it adds some overhead so I
    # don't think it pays off. The member variables are also sorted in
    # the classes (Product and Sum) so the list 'variables' is
    # probably already sorted.
    key = tuple(variables)
    if key in _sum_cache:
        return _sum_cache[key]
    s = Sum(key)
    _sum_cache[key] = s
    return s


_fraction_cache = {}


def create_fraction(num, denom):
    key = (num, denom)
    if key in _fraction_cache:
        return _fraction_cache[key]
    fraction = Fraction(num, denom)
    _fraction_cache[key] = fraction
    return fraction


def generate_aux_constants(constant_decl, name, var_type, print_ops=False):
    "A helper tool to generate code for constant declarations."
    format_comment = format["comment"]
    code = []
    append = code.append
    ops = 0
    for num, expr in sorted((v, k) for k, v in sorted_by_key(constant_decl)):
        # Expand and reduce expression (If we don't already get
        # reduced expressions.)
        expr = expr.expand().reduce_ops()
        if print_ops:
            op = expr.ops()
            ops += op
            append(format_comment("Number of operations: %d" % op))
            append(var_type(name(num), str(expr)))
            append("")
        else:
            ops += expr.ops()
            append(var_type(name(num), str(expr)))

    return (ops, code)


def optimise_code(expr, ip_consts, geo_consts, trans_set):
    """Optimise a given expression with respect to, basis functions,
    integration points variables and geometric constants.  The
    function will update the dictionaries ip_const and geo_consts with
    new declarations and update the trans_set (used
    transformations).

    """

    format_G = format["geometry constant"]
    format_I = format["ip constant"]
    trans_set_update = trans_set.update

    # Return constant symbol if expanded value is zero.
    exp_expr = expr.expand()
    if exp_expr.val == 0.0:
        return create_float(0)

    # Reduce expression with respect to basis function variable.
    basis_expressions = exp_expr.reduce_vartype(BASIS)

    # If we had a product instance we'll get a tuple back so embed in
    # list.
    if not isinstance(basis_expressions, list):
        basis_expressions = [basis_expressions]

    basis_vals = []
    # Process each instance of basis functions.
    for basis, ip_expr in basis_expressions:
        # Get the basis and the ip expression.

        # If we have no basis (like functionals) create a const.
        if not basis:
            basis = create_float(1)

        # If the ip expression doesn't contain any operations skip
        # remainder
        if not ip_expr or ip_expr.val == 0.0:
            basis_vals.append(basis)
            continue
        if not ip_expr.ops() > 0:
            basis_vals.append(create_product([basis, ip_expr]))
            continue

        # Reduce the ip expressions with respect to IP variables.
        ip_expressions = ip_expr.expand().reduce_vartype(IP)

        # If we had a product instance we'll get a tuple back so embed in list.
        if not isinstance(ip_expressions, list):
            ip_expressions = [ip_expressions]

        ip_vals = []
        # Loop ip expressions.
        for ip in sorted(ip_expressions):
            ip_dec, geo = ip

            # Update transformation set with those values that might
            # be embedded in IP terms.
            if ip_dec and ip_dec.val != 0.0:
                trans_set_update([str(x) for x in ip_dec.get_unique_vars(GEO)])

            # Append and continue if we did not have any geo values.
            if not geo or geo.val == 0.0:
                if ip_dec and ip_dec.val != 0.0:
                    ip_vals.append(ip_dec)
                continue

            # Update the transformation set with the variables in the
            # geo term.
            trans_set_update([str(x) for x in geo.get_unique_vars(GEO)])

            # Only declare auxiliary geo terms if we can save
            # operations.
            if geo.ops() > 0:
                # If the geo term is not in the dictionary append it.
                if geo not in geo_consts:
                    geo_consts[geo] = len(geo_consts)

                # Substitute geometry expression.
                geo = create_symbol(format_G(geo_consts[geo]), GEO)

            # If we did not have any ip_declarations use geo, else
            # create a product and append to the list of ip_values.
            if not ip_dec or ip_dec.val == 0.0:
                ip_dec = geo
            else:
                ip_dec = create_product([ip_dec, geo])
            ip_vals.append(ip_dec)

        # Create sum of ip expressions to multiply by basis.
        if len(ip_vals) > 1:
            ip_expr = create_sum(ip_vals)
        elif ip_vals:
            ip_expr = ip_vals.pop()

        # If we can save operations by declaring it as a constant do
        # so, if it is not in IP dictionary, add it and use new name.
        if ip_expr.ops() > 0 and ip_expr.val != 0.0:
            if ip_expr not in ip_consts:
                ip_consts[ip_expr] = len(ip_consts)

            # Substitute ip expression.
            ip_expr = create_symbol(format_I(ip_consts[ip_expr]), IP)

        # Multiply by basis and append to basis vals.
        basis_vals.append(create_product([basis, ip_expr]))

    # Return (possible) sum of basis values.
    if len(basis_vals) > 1:
        return create_sum(basis_vals)
    elif basis_vals:
        return basis_vals[0]

    # Where did the values go?
    error("Values disappeared.")


from .floatvalue import FloatValue
from .symbol import Symbol
from .product import Product
from .sumobj import Sum
from .fraction import Fraction



# -*- coding: utf-8 -*-
"Reordering of entries in reference tensor for interior facets."

# Copyright (C) 2006-2009 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2006-12-01
# Last changed: 2010-02-10

# Python modules
import numpy

# FFC tensor representation modules
from .monomialtransformation import MonomialIndex
from .multiindex import MultiIndex


def reorder_entries(terms):
    """Reorder entries to compute the reference tensor for an interior
    facet from the the reduced reference tensor."""

    # Iterate over terms
    for term in terms:

        # Compute restrictions corresponding to indices
        (restrictions, idims, adims) = __compute_restrictions(term)
        dims = idims + adims

        # Compute position where to insert
        position = []
        for i in range(len(restrictions)):
            dim = dims[i]
            if restrictions[i] == "+":
                position = position + [slice(0, dim // 2)]
            elif restrictions[i] == "-":
                position = position + [slice(dim // 2, dim)]
            else:
                position = position + [slice(0, dim)]

        # Initialize empty reference tensor of double size in each dimension
        tensor = numpy.zeros(dims, dtype=numpy.float)

        # Insert reduced reference tensor into reference tensor
        (A0, GK, optimized_contraction) = term
        tensor[position] = A0.A0
        A0.A0 = tensor

        # Reinitialize indices to new size
        A0.primary_multi_index = MultiIndex([list(range(idim)) for idim in idims])
        A0.secondary_multi_index = MultiIndex([list(range(adim)) for adim in adims])
        GK.secondary_multi_index = A0.secondary_multi_index


def __compute_restrictions(term):
    """Compute restrictions corresponding to indices for given
    term. For indices at basis functions, we need to double the
    size of the reference tensor in the corresponding dimension,
    but for other dimensions corresponding to component indices
    and derivatives, the size remains the same."""

    # Get dimensions for primary and secondary indices
    A0, GK, optimized_contraction = term
    idims = A0.primary_multi_index.dims
    adims = A0.secondary_multi_index.dims

    # Get basis functions for term
    arguments = A0.monomial.arguments

    # Create empty list of restrictions for indices
    restrictions = [None for i in range(len(idims) + len(adims))]

    # Extract restrictions corresponding to primary indices at basis functions
    for i in range(len(idims)):
        for v in arguments:
            if v.index.index_type == MonomialIndex.PRIMARY and v.index.index_id == i:
                restrictions[i] = v.restriction
                break

    # Extract restrictions corresponding to secondary indices at basis functions
    for i in range(len(adims)):
        for v in arguments:
            if v.index.index_type == MonomialIndex.SECONDARY and v.index.index_id == i:
                restrictions[len(idims) + i] = v.restriction
                break

    # Compute new dimensions
    new_idims = [i for i in idims]
    new_adims = [i for i in adims]
    for i in range(len(new_idims)):
        if restrictions[i] is not None:
            new_idims[i] = 2 * new_idims[i]
    for i in range(len(new_adims)):
        if restrictions[i + len(new_idims)] is not None:
            new_adims[i] = 2 * new_adims[i]

    return (restrictions, new_idims, new_adims)



# -*- coding: utf-8 -*-
"Code generator for quadrature representation."

# Copyright (C) 2009-2014 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Mehdi Nikbakht 2010
# Modified by Anders Logg 2013-2014
# Modified by Martin Sandve Alnæs 2013-2014

# Python modules
import functools
import numpy

# UFL modules
from ufl.utils.sorting import sorted_by_key
from ufl.utils.derivativetuples import compute_derivative_tuples
from ufl import custom_integral_types

# FFC modules
from ffc.log import ffc_assert, error, warning
from ffc.cpp import format, remove_unused

from ffc.representationutils import initialize_integral_code

# Utility and optimization functions for quadraturegenerator
from ffc.quadrature.symbolics import generate_aux_constants


def generate_integral_code(ir, prefix, parameters):
    "Generate code for integral from intermediate representation."

    # Prefetch formatting to speedup code generation (well...)
    ret = format["return"]

    # Generate code
    code = initialize_integral_code(ir, prefix, parameters)
    code["num_cells"] = ret(ir["num_cells"])
    code["tabulate_tensor"] = _tabulate_tensor(ir, prefix, parameters)
    code["additional_includes_set"] = ir["additional_includes_set"]

    precision = ir["integrals_metadata"].get("precision")
    if precision is not None and precision != parameters["precision"]:
        warning("Ignoring precision in integral metadata compiled "
                "using quadrature representation. Not implemented.")

    return code


def _tabulate_tensor(ir, prefix, parameters):
    "Generate code for a single integral (tabulate_tensor())."

    # Prefetch formatting to speedup code generation
    f_comment = format["comment"]
    f_G = format["geometry constant"]
    f_const_double = format["assign"]
    f_switch = format["switch"]
    f_float = format["float"]
    f_assign = format["assign"]
    f_A = format["element tensor"]
    f_r = format["free indices"][0]
    f_loop = format["generate loop"]
    f_int = format["int"]
    f_facet = format["facet"]

    # Get data
    opt_par = ir["optimise_parameters"]
    integral_type = ir["integral_type"]
    gdim = ir["geometric_dimension"]
    tdim = ir["topological_dimension"]
    num_facets = ir["num_facets"]
    num_vertices = ir["num_vertices"]
    prim_idims = ir["prim_idims"]
    integrals = ir["trans_integrals"]
    geo_consts = ir["geo_consts"]
    oriented = ir["needs_oriented"]
    element_data = ir["element_data"]
    num_cells = ir["num_cells"]

    # Create sets of used variables
    used_weights = set()
    used_psi_tables = set()
    used_nzcs = set()
    trans_set = set()
    sets = [used_weights, used_psi_tables, used_nzcs, trans_set]

    affine_tables = {}  # TODO: This is not populated anywhere, remove?
    quadrature_rules = ir["quadrature_rules"]

    operations = []
    if integral_type == "cell":

        # Generate code for computing element tensor
        tensor_code, mem_code, num_ops = _generate_element_tensor(integrals,
                                                                  sets,
                                                                  opt_par,
                                                                  gdim,
                                                                  tdim)
        tensor_code = "\n".join(tensor_code)

        # Set operations equal to num_ops (for printing info on operations).
        operations.append([num_ops])

        # Generate code for basic geometric quantities
        jacobi_code = ""
        jacobi_code += format["compute_jacobian"](tdim, gdim)
        jacobi_code += "\n"
        jacobi_code += format["compute_jacobian_inverse"](tdim, gdim)
        if oriented:
            jacobi_code += format["orientation"](tdim, gdim)
        jacobi_code += "\n"
        jacobi_code += format["scale factor snippet"]

        # Generate code for cell volume and circumradius
        jacobi_code += "\n\n" + format["generate cell volume"](tdim, gdim,
                                                               integral_type)
        jacobi_code += "\n\n" + format["generate circumradius"](tdim, gdim,
                                                                integral_type)

    elif integral_type == "exterior_facet":

        # Iterate over facets
        cases = [None for i in range(num_facets)]
        for i in range(num_facets):
            # Update transformer with facets and generate case code +
            # set of used geometry terms.
            c, mem_code, ops = _generate_element_tensor(integrals[i], sets,
                                                        opt_par, gdim, tdim)
            case = [f_comment("Total number of operations to compute element tensor (from this point): %d" % ops)]
            case += c
            cases[i] = "\n".join(case)

            # Save number of operations (for printing info on
            # operations).
            operations.append([i, ops])

        # Generate tensor code for all cases using a switch.
        tensor_code = f_switch(f_facet(None), cases)

        # Generate code for basic geometric quantities
        jacobi_code = ""
        jacobi_code += format["compute_jacobian"](tdim, gdim)
        jacobi_code += "\n"
        jacobi_code += format["compute_jacobian_inverse"](tdim, gdim)
        if oriented:
            jacobi_code += format["orientation"](tdim, gdim)
        jacobi_code += "\n"
        jacobi_code += "\n\n" + format["facet determinant"](tdim, gdim)
        jacobi_code += "\n\n" + format["generate normal"](tdim, gdim,
                                                          integral_type)
        jacobi_code += "\n\n" + format["generate facet area"](tdim, gdim)
        if tdim == 3:
            jacobi_code += "\n\n" + format["generate min facet edge length"](tdim, gdim)
            jacobi_code += "\n\n" + format["generate max facet edge length"](tdim, gdim)

        # Generate code for cell volume and circumradius
        jacobi_code += "\n\n" + format["generate cell volume"](tdim, gdim,
                                                               integral_type)
        jacobi_code += "\n\n" + format["generate circumradius"](tdim, gdim,
                                                                integral_type)

    elif integral_type == "interior_facet":

        # Modify the dimensions of the primary indices because we have
        # a macro element
        prim_idims = [d * 2 for d in prim_idims]

        # Iterate over combinations of facets
        cases = [[None for j in range(num_facets)] for i in range(num_facets)]
        for i in range(num_facets):
            for j in range(num_facets):
                # Update transformer with facets and generate case
                # code + set of used geometry terms.
                c, mem_code, ops = _generate_element_tensor(integrals[i][j],
                                                            sets, opt_par,
                                                            gdim, tdim)
                case = [f_comment("Total number of operations to compute element tensor (from this point): %d" % ops)]
                case += c
                cases[i][j] = "\n".join(case)

                # Save number of operations (for printing info on
                # operations).
                operations.append([i, j, ops])

        # Generate tensor code for all cases using a switch.
        tensor_code = f_switch(f_facet("+"),
                               [f_switch(f_facet("-"),
                                         cases[i]) for i in range(len(cases))])

        # Generate code for basic geometric quantities
        jacobi_code = ""
        for _r in ["+", "-"]:
            jacobi_code += format["compute_jacobian"](tdim, gdim, r=_r)
            jacobi_code += "\n"
            jacobi_code += format["compute_jacobian_inverse"](tdim, gdim, r=_r)
            if oriented:
                jacobi_code += format["orientation"](tdim, gdim, r=_r)
            jacobi_code += "\n"
        jacobi_code += "\n\n" + format["facet determinant"](tdim, gdim, r="+")
        jacobi_code += "\n\n" + format["generate normal"](tdim, gdim,
                                                          integral_type)
        jacobi_code += "\n\n" + format["generate facet area"](tdim, gdim)
        if tdim == 3:
            jacobi_code += "\n\n" + format["generate min facet edge length"](tdim, gdim, r="+")
            jacobi_code += "\n\n" + format["generate max facet edge length"](tdim, gdim, r="+")

        # Generate code for cell volume and circumradius
        jacobi_code += "\n\n" + format["generate cell volume"](tdim, gdim,
                                                               integral_type)
        jacobi_code += "\n\n" + format["generate circumradius"](tdim, gdim,
                                                                integral_type)

    elif integral_type == "vertex":

        # Iterate over vertices
        cases = [None for i in range(num_vertices)]
        for i in range(num_vertices):
            # Update transformer with vertices and generate case code
            # + set of used geometry terms.
            c, mem_code, ops = _generate_element_tensor(integrals[i],
                                                        sets,
                                                        opt_par,
                                                        gdim,
                                                        tdim)
            case = [f_comment("Total number of operations to compute element tensor (from this point): %d" % ops)]
            case += c
            cases[i] = "\n".join(case)

            # Save number of operations (for printing info on
            # operations).
            operations.append([i, ops])

        # Generate tensor code for all cases using a switch.
        tensor_code = f_switch(format["vertex"], cases)

        # Generate code for basic geometric quantities
        jacobi_code = ""
        jacobi_code += format["compute_jacobian"](tdim, gdim)
        jacobi_code += "\n"
        jacobi_code += format["compute_jacobian_inverse"](tdim, gdim)
        if oriented:
            jacobi_code += format["orientation"](tdim, gdim)
        jacobi_code += "\n"
        jacobi_code += "\n\n" + format["facet determinant"](tdim, gdim)  # FIXME: This is not defined in a point???

    elif integral_type in custom_integral_types:

        # Set number of cells
        if integral_type == "cutcell":
            num_cells = 1
        elif integral_type == "interface":
            num_cells = 2
        elif integral_type == "overlap":
            num_cells = 2

        # Warning that more than two cells in only partly supported.
        # The missing piece is to couple multiple cells to
        # restrictions other than '+' and '-'.
        if num_cells > 2:
            warning("Custom integrals with more than two cells only partly supported.")

        # Modify the dimensions of the primary indices because we have a macro element
        if num_cells == 2:
            prim_idims = [d * 2 for d in prim_idims]

        # Check whether we need to generate facet normals
        generate_custom_facet_normal = num_cells == 2

        # Generate code for computing element tensor
        tensor_code, mem_code, num_ops = _generate_element_tensor(integrals,
                                                                  sets,
                                                                  opt_par,
                                                                  gdim,
                                                                  tdim,
                                                                  generate_custom_facet_normal)

        tensor_code = "\n".join(tensor_code)

        # Set operations equal to num_ops (for printing info on
        # operations).
        operations.append([num_ops])

        # FIXME: Jacobi code is only needed when we use cell volume or
        # circumradius.
        # FIXME: Does not seem to be removed by removed_unused.

        # Generate code for basic geometric quantities
        jacobi_code = ""
        for i in range(num_cells):
            r = i if num_cells > 1 else None
            jacobi_code += "\n"
            jacobi_code += f_comment("--- Compute geometric quantities on cell %d ---" % i)
            jacobi_code += "\n\n"
            if num_cells > 1:
                jacobi_code += f_comment("Extract vertex coordinates\n")
                jacobi_code += format["extract_cell_coordinates"]((tdim + 1) * gdim * i, r=i)
                jacobi_code += "\n\n"
            jacobi_code += format["compute_jacobian"](tdim, gdim, r=r)
            jacobi_code += "\n"
            jacobi_code += format["compute_jacobian_inverse"](tdim, gdim, r=r)
            jacobi_code += "\n"
            jacobi_code += format["generate cell volume"](tdim, gdim,
                                                          integral_type,
                                                          r=r if num_cells > 1 else None)
            jacobi_code += "\n"
            jacobi_code += format["generate circumradius"](tdim, gdim,
                                                           integral_type,
                                                           r=r if num_cells > 1 else None)
            jacobi_code += "\n"

    else:
        error("Unhandled integral type: " + str(integral_type))

    # After we have generated the element code for all facets we can
    # remove the unused transformations.
    common = [remove_unused(jacobi_code, trans_set)]

    # FIXME: After introduction of custom integrals, the common code
    # here is not really common anymore. Think about how to
    # restructure this function.

    # Add common code except for custom integrals
    if integral_type not in custom_integral_types:
        common += _tabulate_weights([quadrature_rules[p] for p in sorted(used_weights)])

        # Add common code for updating tables
        name_map = ir["name_map"]
        tables = ir["unique_tables"]
        tables.update(affine_tables)  # TODO: This is not populated anywhere, remove?
        common += _tabulate_psis(tables, used_psi_tables, name_map, used_nzcs,
                                 opt_par, integral_type, gdim)

    # Add special tabulation code for custom integral
    else:
        common += _evaluate_basis_at_quadrature_points(used_psi_tables,
                                                       gdim,
                                                       element_data,
                                                       prefix,
                                                       num_vertices,
                                                       num_cells)

    # Reset the element tensor (array 'A' given as argument to
    # tabulate_tensor() by assembler)
    # Handle functionals.
    common += [f_comment("Reset values in the element tensor.")]
    if prim_idims == []:
        common += [f_assign(f_A(f_int(0)), f_float(0))]
    else:
        dim = functools.reduce(lambda v, u: v * u, prim_idims)
        common += f_loop([f_assign(f_A(f_r), f_float(0))], [(f_r, 0, dim)])

    # Create the constant geometry declarations (only generated if
    # simplify expressions are enabled).
    geo_ops, geo_code = generate_aux_constants(geo_consts, f_G, f_const_double)
    if geo_code:
        common += [f_comment("Number of operations to compute geometry constants: %d." % geo_ops)]
        common += [format["declaration"](format["float declaration"], f_G(len(geo_consts)))]
        common += geo_code

    # Add comments.
    common += ["", f_comment("Compute element tensor using UFL quadrature representation")]
    common += [f_comment("Optimisations: %s" % ", ".join([str((k, opt_par[k]))
                                                          for k in sorted(opt_par.keys())]))]

    for ops in operations:
        # Add geo ops count to integral ops count for writing info.
        if isinstance(ops[-1], int):
            ops[-1] += geo_ops

    return "\n".join(common) + "\n" + tensor_code


def _generate_element_tensor(integrals, sets, optimise_parameters, gdim, tdim,
                             generate_custom_facet_normal=False):
    "Construct quadrature code for element tensors."

    # Prefetch formats to speed up code generation.
    f_comment = format["comment"]
    f_ip = format["integration points"]
    f_I = format["ip constant"]
    f_loop = format["generate loop"]
    f_ip_coords = format["generate ip coordinates"]
    f_coords = format["coordinate_dofs"]
    f_double = format["float declaration"]
    f_decl = format["declaration"]
    f_X = format["ip coordinates"]
    f_C = format["conditional"]

    # Initialise return values.
    element_code = []
    tensor_ops_count = 0

    # TODO: KBO: The members_code was used when I generated the
    # load_table.h file which could load tables of basisfunction. This
    # feature has not been reimplemented. However, with the new design
    # where we only tabulate unique tables (and only non-zero entries)
    # it doesn't seem to be necessary. Should it be deleted?
    members_code = ""

    # We receive a dictionary {num_points: form,}.
    # Loop points and forms.
    for points, terms, functions, ip_consts, coordinate, conditionals in integrals:

        element_code += ["", f_comment("Loop quadrature points for integral.")]

        ip_code = []
        num_ops = 0

        # Generate code to compute coordinates if used.
        if coordinate:
            name, gdim, ip, r = coordinate
            element_code += ["", f_comment("Declare array to hold physical coordinate of quadrature point.")]
            element_code += [f_decl(f_double, f_X(points, gdim))]
            ops, coord_code = f_ip_coords(gdim, tdim, points, name, ip, r)
            ip_code += ["", f_comment("Compute physical coordinate of quadrature point, operations: %d." % ops)]
            ip_code += [coord_code]
            num_ops += ops
            # Update used psi tables and transformation set.
            sets[1].add(name)
            sets[3].add(f_coords(r))

        # Generate code to compute function values.
        if functions:
            func_code, ops = _generate_functions(functions, sets)
            ip_code += func_code
            num_ops += ops

        # Generate code to compute conditionals (might depend on coordinates
        # and function values so put here).
        # TODO: Some conditionals might only depend on geometry so they
        # should be moved outside if possible.
        if conditionals:
            ip_code += [f_decl(f_double, f_C(len(conditionals)))]
            # Sort conditionals (need to in case of nested conditionals).
            reversed_conds = dict([(n, (o, e)) for e, (t, o, n) in conditionals.items()])
            for num in range(len(conditionals)):
                name = format["conditional"](num)
                ops, expr = reversed_conds[num]
                ip_code += [f_comment("Compute conditional, operations: %d." % ops)]
                ip_code += [format["assign"](name, expr)]
                num_ops += ops

        # Generate code for ip constant declarations.
        ip_const_ops, ip_const_code = generate_aux_constants(ip_consts, f_I,
                                                             format["assign"], True)
        num_ops += ip_const_ops
        if ip_const_code:
            ip_code += ["", f_comment("Number of operations to compute ip constants: %d" % ip_const_ops)]
            ip_code += [format["declaration"](format["float declaration"], f_I(len(ip_consts)))]
            ip_code += ip_const_code

        # Generate code to evaluate the element tensor.
        integral_code, ops = _generate_integral_code(points, terms, sets,
                                                     optimise_parameters)
        num_ops += ops
        if points is None:
            quadrature_ops = "unknown"
            tensor_ops_count = "unknown"
        else:
            quadrature_ops = num_ops * points
            tensor_ops_count += quadrature_ops
        ip_code += integral_code
        element_code.append(f_comment
                            ("Number of operations to compute element tensor for following IP loop = %s" % str(quadrature_ops)))

        # Generate code for custom facet normal if necessary
        if generate_custom_facet_normal:
            for line in ip_code:
                if "n_00" in line:
                    ip_code = [format["facet_normal_custom"](gdim)] + ip_code
                    break

        # Loop code over all IPs.
        if points == 0:
            element_code.append(f_comment("Only 1 integration point, omitting IP loop."))
            element_code += ip_code
        elif points is None:
            num_points = "num_quadrature_points"
            element_code += f_loop(ip_code, [(f_ip, 0, num_points)])
        else:
            element_code += f_loop(ip_code, [(f_ip, 0, points)])

    return (element_code, members_code, tensor_ops_count)


def _generate_functions(functions, sets):
    "Generate declarations for functions and code to compute values."

    f_comment = format["comment"]
    f_double = format["float declaration"]
    f_F = format["function value"]
    f_float = format["floating point"]
    f_decl = format["declaration"]
    f_r = format["free indices"][0]
    f_iadd = format["iadd"]
    f_loop = format["generate loop"]

    # Create the function declarations.
    code = ["", f_comment("Coefficient declarations.")]
    code += [f_decl(f_double, f_F(n), f_float(0)) for n in range(len(functions))]

    # Get sets.
    used_psi_tables = sets[1]
    used_nzcs = sets[2]

    # Sort functions after loop ranges.
    function_list = {}
    for key, val in functions.items():
        if val[1] in function_list:
            function_list[val[1]].append(key)
        else:
            function_list[val[1]] = [key]

    total_ops = 0
    # Loop ranges and get list of functions.
    for loop_range, list_of_functions in sorted(function_list.items()):
        function_expr = {}
        # Loop functions.
        func_ops = 0
        for function in list_of_functions:
            # Get name and number.
            number, range_i, ops, psi_name, u_nzcs, ufl_element = functions[function]

            # Add name to used psi names and non zeros name to used_nzcs.
            used_psi_tables.add(psi_name)
            used_nzcs.update(u_nzcs)

            # TODO: This check can be removed for speed later.
            ffc_assert(number not in function_expr, "This is definitely not supposed to happen!")

            # Convert function (that might be a symbol) to a simple
            # string and save.
            function = str(function)
            function_expr[number] = function

            # Get number of operations to compute entry and add to
            # function operations count.
            func_ops += (ops + 1) * range_i

        # Add function operations to total count
        total_ops += func_ops
        code += ["", f_comment("Total number of operations to compute function values = %d" % func_ops)]

        # Sort the functions according to name and create loop to
        # compute the function values.
        lines = [f_iadd(f_F(n), function_expr[n]) for n in sorted(function_expr.keys())]
        code += f_loop(lines, [(f_r, 0, loop_range)])  # TODO: If loop_range == 1, this loop may be unneccessary. Not sure if it's safe to just skip it.

    return code, total_ops


def _generate_integral_code(points, terms, sets, optimise_parameters):
    "Generate code to evaluate the element tensor."

    # Prefetch formats to speed up code generation.
    f_comment = format["comment"]
    f_iadd = format["iadd"]
    f_A = format["element tensor"]
    f_loop = format["generate loop"]
    f_B = format["basis constant"]

    # Initialise return values.
    code = []
    num_ops = 0
    loops = {}

    # Extract sets.
    used_weights, used_psi_tables, used_nzcs, trans_set = sets

    # Loop terms and create code.
    for loop, (data, entry_vals) in sorted(terms.items()):

        # If we don't have any entry values, there's no need to
        # generate the loop.
        if not entry_vals:
            continue

        # Get data.
        t_set, u_weights, u_psi_tables, u_nzcs, basis_consts = data

        # If we have a value, then we also need to update the sets of
        # used variables.
        trans_set.update(t_set)
        used_weights.update(u_weights)
        used_psi_tables.update(u_psi_tables)
        used_nzcs.update(u_nzcs)

        # Generate code for basis constant declarations.
        basis_const_ops, basis_const_code = generate_aux_constants(basis_consts,
                                                                   f_B,
                                                                   format["assign"], True)
        decl_code = []
        if basis_consts:
            decl_code = [format["declaration"](format["float declaration"],
                                               f_B(len(basis_consts)))]
        loops[loop] = [basis_const_ops, decl_code + basis_const_code]

        for entry, value, ops in entry_vals:
            # Compute number of operations to compute entry (add 1
            # because of += in assignment).
            entry_ops = ops + 1

            # Create comment for number of operations
            entry_ops_comment = f_comment("Number of operations to compute entry: %d" % entry_ops)
            entry_code = f_iadd(f_A(entry), value)
            loops[loop][0] += entry_ops
            loops[loop][1] += [entry_ops_comment, entry_code]

    # Write all the loops of basis functions.
    for loop, ops_lines in sorted(loops.items()):
        ops, lines = ops_lines
        prim_ops = functools.reduce(lambda i, j: i * j, [ops] + [l[2] for l in loop])
        # Add number of operations for current loop to total count.
        num_ops += prim_ops
        code += ["", f_comment("Number of operations for primary indices: %d" % prim_ops)]
        code += f_loop(lines, loop)

    return code, num_ops


def _tabulate_weights(quadrature_rules):
    "Generate table of quadrature weights."

    # Prefetch formats to speed up code generation.
    f_float = format["floating point"]
    f_table = format["static const float declaration"]
    f_sep = format["list separator"]
    f_weight = format["weight"]
    f_component = format["component"]
    f_group = format["grouping"]
    f_decl = format["declaration"]
    f_tensor = format["tabulate tensor"]
    f_comment = format["comment"]
    f_int = format["int"]

    code = ["", f_comment("Array of quadrature weights.")]

    # Loop tables of weights and create code.
    for weights, points in quadrature_rules:
        # FIXME: For now, raise error if we don't have weights.
        # We might want to change this later.
        ffc_assert(weights.any(), "No weights.")

        # Create name and value for weight.
        num_points = len(points)
        name = f_weight(num_points)
        value = f_float(weights[0])
        if len(weights) > 1:
            name += f_component("", f_int(num_points))
            value = f_tensor(weights)
        code += [f_decl(f_table, name, value)]

        # Tabulate the quadrature points (uncomment for different
        # parameters).  1) Tabulate the points as: p0, p1, p2, with p0
        # = (x0, y0, z0) etc.  Use f_float to format the value (enable
        # variable precision).
        formatted_points = [f_group(f_sep.join([f_float(val) for val in point]))
                            for point in points]

        # Create comment.
        comment = "Quadrature points on the UFC reference element: " \
                  + f_sep.join(formatted_points)
        code += [f_comment(comment)]

        # 2) Tabulate the coordinates of the points p0, p1, p2 etc.
        #    X: x0, x1, x2
        #    Y: y0, y1, y2
        #    Z: z0, z1, z2
#            comment = "Quadrature coordinates on the UFC reference element: "
#            code += [format["comment"](comment)]

# All points have the same number of coordinates.
#            num_coord = len(points[0])

# All points have x-coordinates.
#            xs = [f_float(p[0]) for p in points]
#            comment = "X: " + f_sep.join(xs)
#            code += [format["comment"](comment)]

#            ys = []
#            zs = []
# Tabulate y-coordinate if we have 2 or more coordinates.
#            if num_coord >= 2:
#                ys = [f_float(p[1]) for p in points]
#                comment = "Y: " + f_sep.join(ys)
#                code += [format["comment"](comment)]
# Only tabulate z-coordinate if we have 3 coordinates.
#            if num_coord == 3:
#                zs = [f_float(p[2]) for p in points]
#                comment = "Z: " + f_sep.join(zs)
#                code += [format["comment"](comment)]

        code += [""]

    return code


def _tabulate_psis(tables, used_psi_tables, inv_name_map, used_nzcs,
                   optimise_parameters, integral_type, gdim):
    "Tabulate values of basis functions and their derivatives at quadrature points."

    # Prefetch formats to speed up code generation.
    f_comment = format["comment"]
    f_table = format["static const float declaration"]
    f_component = format["component"]
    f_const_uint = format["static const uint declaration"]
    f_nzcolumns = format["nonzero columns"]
    f_list = format["list"]
    f_decl = format["declaration"]
    f_tensor = format["tabulate tensor"]
    f_new_line = format["new line"]
    f_int = format["int"]

    # FIXME: Check if we can simplify the tabulation
    code = []
    code += [f_comment("Values of basis functions at quadrature points.")]

    # Get list of non zero columns, if we ignore ones, ignore columns
    # with one component.
    if optimise_parameters["ignore ones"]:
        nzcs = []
        for key, val in sorted(inv_name_map.items()):
            # Check if we have a table of ones or if number of
            # non-zero columns is larger than one.
            if val[1] and len(val[1][1]) > 1 or not val[3]:
                nzcs.append(val[1])
    else:
        nzcs = [val[1] for key, val in sorted(inv_name_map.items())
                if val[1]]

    # TODO: Do we get arrays that are not unique?
    new_nzcs = []
    for nz in nzcs:
        # Only get unique arrays.
        if nz not in new_nzcs:
            new_nzcs.append(nz)

    # Construct name map.
    name_map = {}
    if inv_name_map:
        for name in sorted(inv_name_map):
            if inv_name_map[name][0] in name_map:
                name_map[inv_name_map[name][0]].append(name)
            else:
                name_map[inv_name_map[name][0]] = [name]

    # Loop items in table and tabulate.
    for name in sorted(used_psi_tables):

        # Only proceed if values are still used (if they're not
        # remapped).
        vals = tables[name]

        if vals is not None:

            # Add declaration to name.
            ip, dofs = numpy.shape(vals)
            decl_name = f_component(name, [ip, dofs])

            # Generate array of values.
            value = f_tensor(vals)
            code += [f_decl(f_table, decl_name, f_new_line + value), ""]

        # Tabulate non-zero indices.
        if optimise_parameters["eliminate zeros"]:
            if name in sorted(name_map):
                for n in name_map[name]:
                    if inv_name_map[n][1] and inv_name_map[n][1] in new_nzcs:
                        i, cols = inv_name_map[n][1]
                        if i not in used_nzcs:
                            continue
                        code += [f_comment("Array of non-zero columns")]
                        value = f_list([f_int(c) for c in list(cols)])
                        name_col = f_component(f_nzcolumns(i), len(cols))
                        code += [f_decl(f_const_uint, name_col, value), ""]

                        # Remove from list of columns.
                        new_nzcs.remove(inv_name_map[n][1])
    return code


def _evaluate_basis_at_quadrature_points(psi_tables, gdim, element_data,
                                         form_prefix, num_vertices, num_cells):
    "Generate code for calling evaluate basis (derivatives) at quadrature points"

    # Prefetch formats to speed up code generation
    f_comment = format["comment"]
    f_declaration = format["declaration"]
    f_static_array = format["static array"]
    f_loop = format["generate loop"]
    f_eval_basis_decl = format["eval_basis_decl"]
    f_eval_basis_init = format["eval_basis_init"]
    f_eval_basis = format["eval_basis"]
    f_eval_basis_copy = format["eval_basis_copy"]
    f_eval_derivs_decl = format["eval_derivs_decl"]
    f_eval_derivs_init = format["eval_derivs_init"]
    f_eval_derivs = format["eval_derivs"]
    f_eval_derivs_copy = format["eval_derivs_copy"]

    code = []

    # Extract prefixes for tables
    prefixes = sorted(set(table.split("_")[0] for table in psi_tables))

    # Use lower case prefix for form name
    form_prefix = form_prefix.lower()

    # The psi_tables used by the quadrature code are for scalar
    # components of specific derivatives, while tabulate_basis_all and
    # tabulate_basis_derivatives_all return data including all
    # possible components and derivatives. We therefore need to
    # iterate over prefixes (= elements), call tabulate_basis_all or
    # tabulate_basis_derivatives all, and then extract the relevant
    # data and fill in the psi_tables. We therefore need to extract
    # for each prefix, which tables need to be filled in.

    # For each unique prefix, check which derivatives and components
    # are used
    used_derivatives_and_components = {}
    for prefix in prefixes:
        used_derivatives_and_components[prefix] = {}
        for table in psi_tables:
            if prefix not in table:
                continue

            # Check for derivative
            if "_D" in table:
                d = table.split("_D")[1].split("_")[0]
                n = sum([int(_d) for _d in d])  # FIXME: Assume at most 9 derivatives...
            else:
                n = 0

            # Check for component
            if "_C" in table:
                c = table.split("_C")[1].split("_")[0]
            else:
                c = None

            # Note that derivative has been used
            if n not in used_derivatives_and_components[prefix]:
                used_derivatives_and_components[prefix][n] = set()
            used_derivatives_and_components[prefix][n].add(c)

    # Generate code for setting quadrature weights
    code += [f_comment("Set quadrature weights")]
    code += [f_declaration("const double*", "W", "quadrature_weights")]
    code += [""]

    # Generate code for calling evaluate_basis_[derivatives_]all
    for prefix in prefixes:

        # Get element data for current element
        counter = int(prefix.split("FE")[1])
        space_dim = element_data[counter]["num_element_dofs"]
        value_size = element_data[counter]["physical_value_size"]
        element_classname = element_data[counter]["classname"]

        # Iterate over derivative orders
        for n, components in sorted_by_key(used_derivatives_and_components[prefix]):
            # components are a set and need to be sorted
            components = sorted(components)

            # Code for evaluate_basis_all (n = 0 which means it's not
            # a derivative)
            if n == 0:

                code += [f_comment("--- Evaluation of basis functions ---")]
                code += [""]

                # Compute variables for code generation
                eval_stride = value_size
                eval_size = space_dim * eval_stride
                table_size = num_cells * space_dim

                # Iterate over components and initialize tables
                for c in components:

                    # Set name of table
                    if c is None:
                        table_name = prefix
                    else:
                        table_name = prefix + "_C%s" % c

                    # Generate code for declaration of table
                    code += [f_comment("Create table %s for basis function values on all cells" % table_name)]
                    code += [f_eval_basis_decl % {"table_name": table_name}]
                    code += [f_eval_basis_init % {"table_name": table_name,
                                                  "table_size": table_size}]
                    code += [""]

                # Iterate over cells in macro element and evaluate basis
                for cell_number in range(num_cells):

                    # Compute variables for code generation
                    eval_name = "%s_values_%d" % (prefix, cell_number)
                    table_offset = cell_number * space_dim
                    vertex_offset = cell_number * num_vertices * gdim

                    # Generate block of code for loop
                    block = []

                    # Generate code for calling evaluate_basis_all
                    block += [f_eval_basis % {"classname": element_classname,
                                              "eval_name": eval_name,
                                              "gdim": gdim,
                                              "vertex_offset": vertex_offset}]

                    # Iterate over components and extract values
                    for c in components:

                        # Set name of table and component offset
                        if c is None:
                            table_name = prefix
                            eval_offset = 0
                        else:
                            table_name = prefix + "_C%s" % c
                            eval_offset = int(c)

                        # Generate code for copying values
                        block += [""]
                        block += [f_eval_basis_copy % {"table_name": table_name,
                                                       "eval_name": eval_name,
                                                       "eval_stride": eval_stride,
                                                       "eval_offset": eval_offset,
                                                       "space_dim": space_dim,
                                                       "table_offset": table_offset}]

                    # Generate code
                    code += [f_comment("Evaluate basis functions on cell %d" % cell_number)]
                    code += [f_static_array("double", eval_name, eval_size)]
                    code += f_loop(block, [("ip", 0, "num_quadrature_points")])
                    code += [""]

            # Code for evaluate_basis_derivatives_all (derivative of degree n > 0)
            else:

                code += [f_comment("--- Evaluation of basis function derivatives of order %d ---" % n)]
                code += [""]

                # FIXME: We extract values for all possible
                # derivatives, even
                # FIXME: if not all are used. (For components, we
                # extract only
                # FIXME: components that are actually used.) This may
                # be optimized
                # FIXME: but the extra cost is likely small.

                # Get derivative tuples
                __, deriv_tuples = compute_derivative_tuples(n, gdim)

                # Generate names for derivatives
                derivs = ["".join(str(_d) for _d in d) for d in deriv_tuples]

                # Compute variables for code generation
                eval_stride = value_size * len(derivs)
                eval_size = space_dim * eval_stride
                table_size = num_cells * space_dim

                # Iterate over derivatives and initialize tables
                seen_derivs = set()
                for d in derivs:

                    # Skip derivative if seen before (d^2/dxdy = d^2/dydx)
                    if d in seen_derivs:
                        continue
                    seen_derivs.add(d)

                    # Iterate over components
                    for c in components:

                        # Set name of table
                        if c is None:
                            table_name = prefix + "_D%s" % d
                        else:
                            table_name = prefix + "_C%s_D%s" % (c, d)

                        # Generate code for declaration of table
                        code += [f_comment("Create table %s for basis function derivatives on all cells" % table_name)]
                        code += [(f_eval_derivs_decl % {"table_name": table_name})]
                        code += [(f_eval_derivs_init % {"table_name": table_name,
                                                        "table_size": table_size})]
                        code += [""]

                # Iterate over cells (in macro element)
                for cell_number in range(num_cells):

                    # Compute variables for code generation
                    eval_name = "%s_dvalues_%d_%d" % (prefix, n, cell_number)
                    table_offset = cell_number * space_dim
                    vertex_offset = cell_number * num_vertices * gdim

                    # Generate block of code for loop
                    block = []

                    # Generate code for calling evaluate_basis_derivatives_all
                    block += [f_eval_derivs % {"classname": element_classname,
                                               "eval_name": eval_name,
                                               "gdim": gdim,
                                               "vertex_offset": vertex_offset,
                                               "n": n}]

                    # Iterate over derivatives and extract values
                    seen_derivs = set()
                    for i, d in enumerate(derivs):

                        # Skip derivative if seen before (d^2/dxdy = d^2/dydx)
                        if d in seen_derivs:
                            continue
                        seen_derivs.add(d)

                        # Iterate over components
                        for c in components:

                            # Set name of table and component offset
                            if c is None:
                                table_name = prefix + "_D%s" % d
                                eval_offset = i
                            else:
                                table_name = prefix + "_C%s_D%s" % (c, d)
                                eval_offset = len(derivs) * int(c) + i

                            # Generate code for copying values
                            block += [""]
                            block += [(f_eval_derivs_copy % {"table_name": table_name,
                                                             "eval_name": eval_name,
                                                             "eval_stride": eval_stride,
                                                             "eval_offset": eval_offset,
                                                             "space_dim": space_dim,
                                                             "table_offset": table_offset})]

                    # Generate code
                    code += [f_comment("Evaluate basis function derivatives on cell %d" % cell_number)]
                    code += [f_static_array("double", eval_name, eval_size)]
                    code += f_loop(block, [("ip", 0, "num_quadrature_points")])
                    code += [""]

                # Add newline
                code += [""]

    return code



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>

"""FFC/UFC specific variable definitions."""

from ufl.corealg.multifunction import MultiFunction
from ufl.measure import custom_integral_types

from ffc.log import error, warning

from ffc.uflacs.backends.ffc.symbols import FFCBackendSymbols
from ffc.uflacs.backends.ffc.common import num_coordinate_component_dofs


class FFCBackendDefinitions(MultiFunction):
    """FFC specific code definitions."""
    def __init__(self, ir, language, symbols, parameters):
        MultiFunction.__init__(self)

        # Store ir and parameters
        self.integral_type = ir["integral_type"]
        self.entitytype = ir["entitytype"]
        self.language = language
        self.symbols = symbols
        self.parameters = parameters

        # TODO: Make this configurable for easy experimentation with dolfin!
        # Coordinate dofs for each component are interleaved? Must match dolfin.
        self.interleaved_components = True # parameters["interleaved_coordinate_component_dofs"]


    # === Generate code to define variables for ufl types ===

    def expr(self, t, mt, tabledata, num_points, access):
        error("Unhandled type {0}".format(type(t)))


    #def quadrature_weight(self, e, mt, tabledata, num_points, access):
    #    "Quadrature weights are precomputed and need no code."
    #    return []


    def constant_value(self, e, mt, tabledata, num_points, access):
        "Constants simply use literals in the target language."
        return []


    def argument(self, t, mt, tabledata, num_points, access):
        "Arguments are accessed through element tables."
        return []


    def coefficient(self, t, mt, tabledata, num_points, access):
        "Return definition code for coefficients."
        L = self.language

        ttype = tabledata.ttype
        begin, end = tabledata.dofrange

        #fe_classname = ir["classnames"]["finite_element"][t.ufl_element()]

        if ttype == "zeros":
            debug("Not expecting zero coefficients to get this far.")
            return []

        # For a constant coefficient we reference the dofs directly, so no definition needed
        if ttype == "ones" and (end - begin) == 1:
            return []

        # For quadrature elements we reference the dofs directly, so no definition needed
        if ttype == "quadrature":
            return []

        assert begin < end

        # Get access to element table
        FE = self.symbols.element_table(tabledata, self.entitytype, mt.restriction)

        unroll = len(tabledata.dofmap) != end - begin
        #unroll = True
        if unroll:
            # TODO: Could also use a generated constant dofmap here like in block code
            # Unrolled loop to accumulate linear combination of dofs and tables
            values = [self.symbols.coefficient_dof_access(mt.terminal, idof) * FE[i]
                      for i, idof in enumerate(tabledata.dofmap)]
            value = L.Sum(values)
            code = [
                L.VariableDecl("const double", access, value)
                ]
        else:
            # Loop to accumulate linear combination of dofs and tables
            ic = self.symbols.coefficient_dof_sum_index()
            dof_access = self.symbols.coefficient_dof_access(mt.terminal, ic + begin)
            code = [
                L.VariableDecl("double", access, 0.0),
                L.ForRange(ic, 0, end - begin,
                           body=[L.AssignAdd(access, dof_access * FE[ic])])
                ]
        return code


    def _define_coordinate_dofs_lincomb(self, e, mt, tabledata, num_points, access):
        "Define x or J as a linear combination of coordinate dofs with given table data."
        L = self.language

        # Get properties of domain
        domain = mt.terminal.ufl_domain()
        #tdim = domain.topological_dimension()
        gdim = domain.geometric_dimension()
        coordinate_element = domain.ufl_coordinate_element()
        #degree = coordinate_element.degree()
        num_scalar_dofs = num_coordinate_component_dofs(coordinate_element)

        # Reference coordinates are known, no coordinate field, so we compute
        # this component as linear combination of coordinate_dofs "dofs" and table

        # Find table name and dof range it corresponds to
        ttype = tabledata.ttype
        begin, end = tabledata.dofrange

        assert end - begin <= num_scalar_dofs
        assert ttype != "zeros"
        assert ttype != "quadrature"
        #xfe_classname = ir["classnames"]["finite_element"][coordinate_element]
        #sfe_classname = ir["classnames"]["finite_element"][coordinate_element.sub_elements()[0]]

        # Get access to element table
        FE = self.symbols.element_table(tabledata, self.entitytype, mt.restriction)

        inline = True

        if ttype == "zeros":
            # Not sure if this will ever happen
            debug("Not expecting zeros for %s." % (e._ufl_class_.__name__,))
            code = [
                L.VariableDecl("const double", access, L.LiteralFloat(0.0))
                ]
        elif ttype == "ones":
            # Not sure if this will ever happen
            debug("Not expecting ones for %s." % (e._ufl_class_.__name__,))
            # Inlined version (we know this is bounded by a small number)
            dof_access = self.symbols.domain_dofs_access(gdim, num_scalar_dofs,
                                                         mt.restriction,
                                                         self.interleaved_components)
            values = [dof_access[idof] for idof in tabledata.dofmap]
            value = L.Sum(values)
            code = [
                L.VariableDecl("const double", access, value)
                ]
        elif inline:
            # Inlined version (we know this is bounded by a small number)
            dof_access = self.symbols.domain_dofs_access(gdim, num_scalar_dofs,
                                                         mt.restriction,
                                                         self.interleaved_components)
            # Inlined loop to accumulate linear combination of dofs and tables
            value = L.Sum([dof_access[idof] * FE[i]
                           for i, idof in enumerate(tabledata.dofmap)])
            code = [
                L.VariableDecl("const double", access, value)
                ]
        else:  # TODO: Make an option to test this version for performance
            # Assuming contiguous dofmap here
            assert len(tabledata.dofmap) == end - begin

            # Generated loop version:
            ic = self.symbols.coefficient_dof_sum_index()
            dof_access = self.symbols.domain_dof_access(ic + begin, mt.flat_component,
                                                        gdim, num_scalar_dofs,
                                                        mt.restriction, self.interleaved_components)

            # Loop to accumulate linear combination of dofs and tables
            code = [
                L.VariableDecl("double", access, 0.0),
                L.ForRange(ic, 0, end - begin,
                           body=[L.AssignAdd(access, dof_access * FE[ic])])
                ]

        return code


    def spatial_coordinate(self, e, mt, tabledata, num_points, access):
        """Return definition code for the physical spatial coordinates.

        If physical coordinates are given:
          No definition needed.

        If reference coordinates are given:
          x = sum_k xdof_k xphi_k(X)

        If reference facet coordinates are given:
          x = sum_k xdof_k xphi_k(Xf)
        """
        if self.integral_type in custom_integral_types:
            # FIXME: Jacobian may need adjustment for custom_integral_types
            if mt.local_derivatives:
                error("FIXME: Jacobian in custom integrals is not implemented.")
            return []
        else:
            return self._define_coordinate_dofs_lincomb(e, mt, tabledata, num_points, access)


    def cell_coordinate(self, e, mt, tabledata, num_points, access):
        """Return definition code for the reference spatial coordinates.

        If reference coordinates are given::

            No definition needed.

        If physical coordinates are given and domain is affine::

            X = K*(x-x0)

        This is inserted symbolically.

        If physical coordinates are given and domain is non- affine::

            Not currently supported.

        """
        # Should be either direct access to points array or symbolically computed
        return []


    def jacobian(self, e, mt, tabledata, num_points, access):
        """Return definition code for the Jacobian of x(X).

        J = sum_k xdof_k grad_X xphi_k(X)
        """
        # TODO: Jacobian may need adjustment for custom_integral_types
        return self._define_coordinate_dofs_lincomb(e, mt, tabledata, num_points, access)


    def cell_orientation(self, e, mt, tabledata, num_points, access):
        # Would be nicer if cell_orientation was a double variable input,
        # but this is how dolfin/ufc/ffc currently passes this information.
        # 0 means up and gives +1.0, 1 means down and gives -1.0.
        L = self.language
        co = self.symbols.cell_orientation_argument(mt.restriction)
        expr = L.Conditional(L.EQ(co, L.LiteralInt(1)),
                             L.LiteralFloat(-1.0), L.LiteralFloat(+1.0))
        code = [
            L.VariableDecl("const double", access, expr)
            ]
        return code


    def _expect_table(self, e, mt, tabledata, num_points, access):
        "These quantities refer to constant tables defined in ufc_geometry.h."
        # TODO: Inject const static table here instead?
        return []
    reference_cell_volume = _expect_table
    reference_facet_volume = _expect_table
    reference_normal = _expect_table
    cell_facet_jacobian = _expect_table
    cell_edge_vectors = _expect_table
    facet_edge_vectors = _expect_table
    facet_orientation = _expect_table


    def _expect_symbolic_lowering(self, e, mt, tabledata, num_points, access):
        "These quantities are expected to be replaced in symbolic preprocessing."
        error("Expecting {0} to be replaced in symbolic preprocessing.".format(type(e)))
    facet_normal = _expect_symbolic_lowering
    cell_normal = _expect_symbolic_lowering
    jacobian_inverse = _expect_symbolic_lowering
    jacobian_determinant = _expect_symbolic_lowering
    facet_jacobian = _expect_symbolic_lowering
    facet_jacobian_inverse = _expect_symbolic_lowering
    facet_jacobian_determinant = _expect_symbolic_lowering



# -*- coding: utf-8 -*-
"This module implements efficient integration of monomial forms."

# Copyright (C) 2004-2011 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Thanks to Robert C. Kirby for suggesting the initial algorithm that
# this implementation is based on.
#
# Modified by Garth N. Wells, 2006
# Modified by Marie E. Rognes, 2008
# Modified by Kristian B. Oelgaard, 2009

# Python modules
import numpy
import time

# FFC modules
from ffc.log import info, debug, error
from ffc.fiatinterface import create_element
from ffc.fiatinterface import map_facet_points
from ffc.representationutils import create_quadrature_points_and_weights

# FFC tensor representation modules
from .multiindex import build_indices
from .monomialextraction import MonomialException
from .monomialtransformation import MonomialIndex


def integrate(monomial,
              integral_type,
              facet0, facet1,
              quadrature_degree,
              quadrature_rule,
              cell):
    """Compute the reference tensor for a given monomial term of a
    multilinear form"""

    info("Precomputing integrals on reference element")

    # Start timing
    tic = time.time()

    # Initialize quadrature points and weights
    (points, weights) = create_quadrature_points_and_weights(
        integral_type, cell, quadrature_degree, quadrature_rule)

    # Initialize quadrature table for basis functions
    table = _init_table(monomial.arguments,
                        integral_type,
                        points,
                        facet0, facet1)

    # Compute table Psi for each factor
    psis = [_compute_psi(v, table, len(points))
            for v in monomial.arguments]

    # Compute product of all Psis
    A0 = _compute_product(psis, monomial.float_value * weights)

    # Report elapsed time and number of entries
    toc = time.time() - tic
    num_entries = numpy.prod(numpy.shape(A0))
    debug("%d entries computed in %.3g seconds" % (num_entries, toc))
    debug("Shape of reference tensor: " + str(numpy.shape(A0)))

    return A0


def _init_table(arguments, integral_type, points, facet0, facet1):
    """Initialize table of basis functions and their derivatives at
    the given quadrature points for each element."""

    # Compute maximum number of derivatives for each element
    num_derivatives = {}
    for v in arguments:
        ufl_element = v.element
        order = len(v.derivatives)
        if ufl_element in num_derivatives:
            num_derivatives[ufl_element] = max(order, num_derivatives[ufl_element])
        else:
            num_derivatives[ufl_element] = order

    # Call FIAT to tabulate the basis functions for each element
    table = {}
    for (ufl_element, order) in num_derivatives.items():
        fiat_element = create_element(ufl_element)
        if integral_type == "cell":
            table[(ufl_element, None)] = fiat_element.tabulate(order, points)
        elif integral_type == "exterior_facet":
            x = map_facet_points(points, facet0)
            table[(ufl_element, None)] = fiat_element.tabulate(order, x)
        elif integral_type == "interior_facet":
            x0 = map_facet_points(points, facet0)
            x1 = map_facet_points(points, facet1)
            table[(ufl_element, "+")] = fiat_element.tabulate(order, x0)
            table[(ufl_element, "-")] = fiat_element.tabulate(order, x1)

    return table


def _compute_psi(v, table, num_points):
    "Compute the table Psi for the given basis function v."

    # We just need to pick the values for Psi from the table, which is
    # somewhat tricky since the table created by tabulate_jet() is a
    # mix of list, dictionary and numpy.array.
    #
    # The dimensions of the resulting table are ordered as follows:
    #
    #     one dimension  corresponding to quadrature points
    #     all dimensions corresponding to internal  Indices
    #     all dimensions corresponding to primary   Indices
    #     all dimensions corresponding to secondary Indices
    #
    # All fixed Indices are removed here. The first set of dimensions
    # corresponding to quadrature points and internal Indices are removed
    # later when we sum over these dimensions.

    # Get topological dimension of cell
    tdim = v.element.cell().topological_dimension()

    # Get indices and shapes for components
    if len(v.components) == 0:
        cindex = []
        cshape = []
    elif len(v.components) == 1:
        cindex = [v.components[0]]
        cshape = [len(v.components[0].index_range)]
    else:
        raise MonomialException("Can only handle rank 0 or rank 1 tensors.")

    # Get indices and shapes for derivatives
    dindex = [d for d in v.derivatives]
    dshape = [len(d.index_range) for d in v.derivatives]

    # Get indices and shapes for basis functions
    vindex = [v.index]
    vshape = [len(v.index.index_range)]

    # Create list of indices that label the dimensions of the tensor Psi
    indices = cindex + dindex + vindex
    shapes = cshape + dshape + vshape + [num_points]

    # Initialize tensor Psi: component, derivatives, basis function, points
    Psi = numpy.zeros(shapes, dtype=numpy.float)

    # Iterate over derivative indices
    dlists = build_indices([index.index_range for index in dindex]) or [[]]
    etable = table[(v.element, v.restriction)]
    if len(cindex) > 0:
        for component in range(len(cindex[0].index_range)):
            for dlist in dlists:
                # Translate derivative multiindex to lookup tuple
                dtuple = _multiindex_to_tuple(dlist, tdim)
                # Get values from table
                Psi[component][tuple(dlist)] = \
                    etable[dtuple][:, cindex[0].index_range[component], :]
    else:
        for dlist in dlists:
            # Translate derivative multiindex to lookup tuple
            dtuple = _multiindex_to_tuple(dlist, tdim)
            # Get values from table
            Psi[tuple(dlist)] = etable[dtuple]

    # Rearrange Indices as (fixed, internal, primary, secondary)
    (rearrangement, num_indices) = _compute_rearrangement(indices)
    indices = [indices[i] for i in rearrangement]
    Psi = numpy.transpose(Psi, rearrangement + (len(indices),))

    # Remove fixed indices
    for i in range(num_indices[0]):
        Psi = Psi[0, ...]
    indices = [index for index in indices
               if not index.index_type == MonomialIndex.FIXED]

    # Put quadrature points first
    rank = Psi.ndim
    Psi = numpy.transpose(Psi, (rank - 1,) + tuple(range(0, rank - 1)))

    # Compute internal index positions for current Psi
    bpart = [i.index_id for i in indices if i.index_type == MonomialIndex.INTERNAL]

    return (Psi, indices, bpart)


def _compute_product(psis, weights):
    "Compute special product of list of Psis."

    # The reference tensor is obtained by summing over quadrature
    # points and internal Indices the outer product of all the Psis
    # with the first dimension (corresponding to quadrature points)
    # and all internal dimensions removed.

    # Initialize zero reference tensor (will be rearranged later)
    (shape, indices) = _compute_shape(psis)
    A0 = numpy.zeros(shape, dtype=numpy.float)

    # Initialize list of internal multiindices
    bshape = _compute_internal_shape(psis)
    bindices = build_indices([list(range(b)) for b in bshape]) or [[]]

    # Sum over quadrature points and internal indices
    num_points = len(weights)
    for q in range(num_points):
        for b in bindices:

            # Compute outer products of subtables for current (q, b)
            B = weights[q]
            for (Psi, index, bpart) in psis:
                B = numpy.multiply.outer(B, Psi[tuple([q] + [b[i] for i in bpart])])

            # Add product to reference tensor
            numpy.add(A0, B, A0)

    # Rearrange Indices as (primary, secondary)
    (rearrangement, num_indices) = _compute_rearrangement(indices)
    A0 = numpy.transpose(A0, rearrangement)

    return A0


def _compute_rearrangement(indices):
    """
    Compute rearrangement tuple for given list of Indices, so that the
    tuple reorders the given list of Indices with fixed, primary,
    secondary and internal Indices in rising order.
    """
    fixed = _find_indices(indices, MonomialIndex.FIXED)
    internal = _find_indices(indices, MonomialIndex.INTERNAL)
    primary = _find_indices(indices, MonomialIndex.PRIMARY)
    secondary = _find_indices(indices, MonomialIndex.SECONDARY)
    assert len(fixed + internal + primary + secondary) == len(indices)
    return (tuple(fixed + internal + primary + secondary),
            (len(fixed), len(internal), len(primary), len(secondary)))


def _compute_shape(psis):
    "Compute shape of reference tensor from given list of tables."
    shape, indices = [], []
    for (Psi, index, bpart) in psis:
        num_internal = len([0 for i in index if i.index_type == MonomialIndex.INTERNAL])
        shape += numpy.shape(Psi)[1 + num_internal:]
        indices += index[num_internal:]
    return (shape, indices)


def _compute_internal_shape(psis):
    """
    Compute shape for internal indices from given list of tables.
    Also compute a list of mappings from each table to the internal
    dimensions associated with that table.
    """
    # First find the number of different internal indices (check maximum)
    bs = [b for (Psi, index, bpart) in psis for b in bpart]
    if len(bs) == 0:
        return []
    bmax = max(bs)
    # Find the dimension for each internal index
    bshape = [0 for i in range(bmax + 1)]
    for (Psi, index, bpart) in psis:
        for i in range(len(bpart)):
            bshape[bpart[i]] = numpy.shape(Psi)[i + 1]
    # Check that we found the shape for each internal index
    if 0 in bshape:
        error("Unable to compute the shape for each internal index.")
    return bshape


def _find_indices(indices, index_type):
    "Return sorted list of positions for given index type."
    pos = [i for i in range(len(indices)) if indices[i].index_type == index_type]
    val = [indices[i].index_id for i in range(len(indices)) if indices[i].index_type == index_type]
    return [pos[i] for i in numpy.argsort(val)]


def _multiindex_to_tuple(dindex, cell_dimension):
    """
    Compute lookup tuple from given derivative multiindex. Necessary
    since the table we get from FIAT is a dictionary with the tuples
    as keys. A derivative tuple specifies the number of derivatives in
    each space dimension, rather than listing the space dimensions for
    the derivatives.
    """
    dtuple = [0 for i in range(cell_dimension)]
    for d in dindex:
        dtuple[d] += 1
    return tuple(dtuple)



# -*- coding: utf-8 -*-
# Based on original implementation by Martin Alnes and Anders Logg
#
# Modified by Anders Logg 2015

__all__ = ["dolfin_tag", "stl_includes", "dolfin_includes", "snippets"]

dolfin_tag = "// DOLFIN wrappers"

stl_includes = """\
// Standard library includes
#include <string>
"""

dolfin_includes = """\
// DOLFIN includes
#include <dolfin/common/NoDeleter.h>
#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/MultiMesh.h>
#include <dolfin/fem/FiniteElement.h>
#include <dolfin/fem/DofMap.h>
#include <dolfin/fem/Form.h>
#include <dolfin/fem/MultiMeshForm.h>
#include <dolfin/function/FunctionSpace.h>
#include <dolfin/function/MultiMeshFunctionSpace.h>
#include <dolfin/function/GenericFunction.h>
#include <dolfin/function/CoefficientAssigner.h>
#include <dolfin/function/MultiMeshCoefficientAssigner.h>
#include <dolfin/adaptivity/ErrorControl.h>
#include <dolfin/adaptivity/GoalFunctional.h>
#include <dolfin/la/GenericVector.h>"""


snippets = {"shared_ptr_space":
            ("std::shared_ptr<const dolfin::FunctionSpace> %s",
             "    _function_spaces[%d] = %s;"),
            "referenced_space":
            ("const dolfin::FunctionSpace& %s",
             "    _function_spaces[%d] = reference_to_no_delete_pointer(%s);"),
            "multimesh_shared_ptr_space":
            ("std::shared_ptr<const dolfin::MultiMeshFunctionSpace> %s",
             None),
            "multimesh_referenced_space":
            ("const dolfin::MultiMeshFunctionSpace& %s",
             None),
            "shared_ptr_mesh":
            ("std::shared_ptr<const dolfin::Mesh> mesh",
             "    _mesh = mesh;"),
            "shared_ptr_multimesh":
                ("std::shared_ptr<const dolfin::MultiMesh> mesh",
                 "    _multimesh = mesh;"),
            "referenced_mesh":
            ("const dolfin::Mesh& mesh",
             "    _mesh = reference_to_no_delete_pointer(mesh);"),
            "shared_ptr_coefficient":
            ("std::shared_ptr<const dolfin::GenericFunction> %s",
             "    this->%s = %s;"),
            "shared_ptr_ref_coefficient":
            ("std::shared_ptr<const dolfin::GenericFunction> %s",
             "    this->%s = *%s;"),
            "referenced_coefficient":
            ("const dolfin::GenericFunction& %s",
             "    this->%s = %s;"),
            "functionspace":
            ("TestSpace", "TrialSpace"),
            "multimeshfunctionspace":
            ("MultiMeshTestSpace", "MultiMeshTrialSpace")
            }
